import glob
import os

import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.corpus.subset import subview
from audiomate.utils import textfile
from . import base


class SpeechCommandsReader(base.CorpusReader):
    """
    Reads the google speech commands dataset.

    .. seealso::

        `Launching Speech Commands DS <https://research.googleblog.com/2017/08/launching-speech-commands-dataset.html>`_
            Blog-Entry on the release of the speech commands dataset.

    """

    @classmethod
    def type(cls):
        return 'speech-commands'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        for folder in os.scandir(path):
            if folder.is_dir() and not folder.name.startswith('_'):
                SpeechCommandsReader._load_folder(folder, corpus)
                SpeechCommandsReader._create_subviews(path, corpus)

        return corpus

    @staticmethod
    def _load_folder(folder_entry, corpus):
        """ Load the given subfolder into the corpus (e.g. bed, one, ...) """
        for wav_path in glob.glob(os.path.join(folder_entry.path, '*.wav')):
            wav_name = os.path.basename(wav_path)
            basename, __ = os.path.splitext(wav_name)

            command = folder_entry.name
            file_idx = '{}_{}'.format(basename, command)
            issuer_idx = str(basename).split('_', maxsplit=1)[0]

            corpus.new_file(wav_path, file_idx)

            if issuer_idx not in corpus.issuers.keys():
                corpus.import_issuers(issuers.Speaker(
                    issuer_idx
                ))

            utt = corpus.new_utterance(file_idx, file_idx, issuer_idx)

            labels = annotations.LabelList.create_single(command, idx=audiomate.corpus.LL_WORD_TRANSCRIPT)
            utt.set_label_list(labels)

    @staticmethod
    def _create_subviews(path, corpus):
        """ Load the subviews based on testing_list.txt and validation_list.txt """
        test_list_path = os.path.join(path, 'testing_list.txt')
        dev_list_path = os.path.join(path, 'validation_list.txt')

        test_list = textfile.read_separated_lines(test_list_path, separator='/', max_columns=2)
        dev_list = textfile.read_separated_lines(dev_list_path, separator='/', max_columns=2)

        test_set = set(['{}_{}'.format(os.path.splitext(x[1])[0], x[0]) for x in test_list])
        dev_set = set(['{}_{}'.format(os.path.splitext(x[1])[0], x[0]) for x in dev_list])
        inv_train_set = test_set.union(dev_set)

        train_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs=inv_train_set, inverse=True)
        train_view = subview.Subview(corpus, filter_criteria=train_filter)
        corpus.import_subview('train', train_view)

        dev_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs=dev_set, inverse=False)
        dev_view = subview.Subview(corpus, filter_criteria=dev_filter)
        corpus.import_subview('dev', dev_view)

        test_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs=test_set, inverse=False)
        test_view = subview.Subview(corpus, filter_criteria=test_filter)
        corpus.import_subview('test', test_view)



"""
The textfile module contains functions for reading and writing textfiles.
"""

import os

from audiomate.utils import text


def read_separated_lines(path, separator=' ', max_columns=-1, keep_empty=False):
    """
    Reads a text file where each line represents a record with some separated columns.

    Parameters:
        path (str): Path to the file to read.
        separator (str): Separator that is used to split the columns.
        max_columns (int): Number of max columns (if the separator occurs within the last column).
        keep_empty (bool): If True empty columns are returned as well.

    Returns:
        list: A list containing a list for each line read.
    """

    gen = read_separated_lines_generator(path, separator, max_columns, keep_empty=keep_empty)
    return list(gen)


def read_separated_lines_with_first_key(path: str, separator: str = ' ', max_columns: int = -1,
                                        keep_empty: bool = False):
    """
    Reads the separated lines of a file and return a dictionary with the first column as keys, value
    is a list with the rest of the columns.

    Parameters:
        path (str): Path to the file to read.
        separator (str): Separator that is used to split the columns.
        max_columns (str): Number of max columns (if the separator occurs within the last column).
        keep_empty (bool): If True empty columns are returned as well.

    Returns:
        dict: Dictionary with list of column values and first column value as key.
    """
    gen = read_separated_lines_generator(path, separator, max_columns, keep_empty=keep_empty)

    dic = {}

    for record in gen:
        if len(record) > 0:
            dic[record[0]] = record[1:len(record)]

    return dic


def read_key_value_lines(path, separator=' ', default_value=''):
    """
    Reads lines of a text file with two columns as key/value dictionary.

    Parameters:
        path (str): Path to the file.
        separator (str): Separator that is used to split key and value.
        default_value (str): If no value is given this value is used.

    Returns:
        dict: A dictionary with first column as key and second as value.
    """
    gen = read_separated_lines_generator(path, separator, 2)

    dic = {}

    for record in gen:
        if len(record) > 1:
            dic[record[0]] = record[1]
        elif len(record) > 0:
            dic[record[0]] = default_value

    return dic


def write_separated_lines(path, values, separator=' ', sort_by_column=0):
    """
    Writes list or dict to file line by line. Dict can have list as value then they written
    separated on the line.

    Parameters:
        path (str): Path to write file to.
        values (dict, list): A dictionary or a list to write to the file.
        separator (str): Separator to use between columns.
        sort_by_column (int): if >= 0, sorts the list by the given index, if its 0 or 1 and its a
                              dictionary it sorts it by either the key (0) or value (1). By default
                              0, meaning sorted by the first column or the key.
    """
    f = open(path, 'w', encoding='utf-8')

    if type(values) is dict:
        if sort_by_column in [0, 1]:
            items = sorted(values.items(), key=lambda t: t[sort_by_column])
        else:
            items = values.items()

        for key, value in items:
            if type(value) in [list, set]:
                value = separator.join([str(x) for x in value])

            f.write('{}{}{}\n'.format(key, separator, value))
    elif type(values) is list or type(values) is set:
        if 0 <= sort_by_column < len(values):
            items = sorted(values)
        else:
            items = values

        for record in items:
            str_values = [str(value) for value in record]

            f.write('{}\n'.format(separator.join(str_values)))

    f.close()


def read_separated_lines_generator(path, separator=' ', max_columns=-1,
                                   ignore_lines_starting_with=[], keep_empty=False):
    """
    Creates a generator through all lines of a file and returns the splitted line.

    Parameters:
        path: Path to the file.
        separator: Separator that is used to split the columns.
        max_columns: Number of max columns (if the separator occurs within the last column).
        ignore_lines_starting_with: Lines starting with a string in this list will be ignored.
        keep_empty (bool): If True empty columns are returned as well.
    """
    if not os.path.isfile(path):
        print('File doesnt exist or is no file: {}'.format(path))
        return

    f = open(path, 'r', errors='ignore', encoding='utf-8')

    if max_columns > -1:
        max_splits = max_columns - 1
    else:
        max_splits = -1

    for line in f:
        if keep_empty:
            stripped_line = line
        else:
            stripped_line = line.strip()

        should_ignore = text.starts_with_prefix_in_list(stripped_line, ignore_lines_starting_with)

        if not should_ignore and stripped_line != '':
            record = stripped_line.split(sep=separator, maxsplit=max_splits)
            record = [field.strip() for field in record]
            yield record

    f.close()



import abc


class FailedDownloadException(Exception):
    pass


class CorpusDownloader(metaclass=abc.ABCMeta):
    """
    Abstract class for downloading a corpus.

    To implement a downloader for a custom format, programmers are expected to subclass this class and to implement all
    abstract methods. The documentation of each abstract methods details the requirements that have to be met by an
    implementation.
    """

    def download(self, target_path):
        """
        Downloads the data of the corpus and saves it to the given path.
        The data has to be saved in a way, so that the corresponding ``CorpusReader`` can load the corpus.

        Args:
            target_path (str): The path to save the data to.
        """
        return self._download(target_path)

    @classmethod
    @abc.abstractmethod
    def type(cls):
        """
        Returns a string that uniquely identifies the downloader. This is usually the name of the corpus, for example
        `musan` or `timit`. Users can use this string to obtain an instance of the desired reader through
        :meth:`~audiomate.corpus.io.create_downloader_of_type` or get a list of all built-in downloaders with
        :meth:`~audiomate.corpus.io.available_downloaders`.

        Returns:
            str: Name of the downloader
        """
        return 'not_implemented'

    @abc.abstractmethod
    def _download(self, target_path):
        """
        Performs the actual downloading of the corpus.

        Args:
            target_path (str): Path to a directory where the data should be saved to.
        """
        pass


class CorpusReader(metaclass=abc.ABCMeta):
    """
    Abstract class for reading a corpus.

    To implement a reader for a custom format, programmers are expected to subclass this class and to implement all
    abstract methods. The documentation of each abstract methods details the requirements that have to be met by an
    implementation.
    """

    def load(self, path):
        """
        Load and return the corpus from the given path.

        Args:
            path (str): Path to the data set to load.

        Returns:
            Corpus: The loaded corpus

        Raises:
            IOError: When the data set is invalid, for example because required files (annotations, …) are missing.
        """
        # Check for missing files
        missing_files = self._check_for_missing_files(path)

        if len(missing_files) > 0:
            raise IOError('Invalid data set of type {}: files {} not found at {}'.format(
                self.type(), ' '.join(missing_files), path))

        return self._load(path)

    @classmethod
    @abc.abstractmethod
    def type(cls):
        """
        Returns a string that uniquely identifies the reader. This is usually the name of the corpus, for example
        `musan` or `timit`. Users can use this string to obtain an instance of the desired reader through
        :meth:`~audiomate.corpus.io.create_reader_of_type` or get a list of all built-in readers with
        :meth:`~audiomate.corpus.io.available_readers`.

        Returns:
            str: Name of the reader
        """
        return 'not_implemented'

    @abc.abstractmethod
    def _load(self, path):
        """
        Performs the actual reading of the corpus.

        Implementations do not have to call :meth:`~audiomate.corpus.io.CorpusReader._check_for_missing_files`
        themselves. This is automatically done by :meth:`~audiomate.corpus.io.CorpusReader.load`.

        Args:
            path (str): Path to a directory where the data set resides.

        Returns:
            Corpus: The loaded corpus
        """
        pass

    @abc.abstractmethod
    def _check_for_missing_files(self, path):
        """
        Tests whether all required files (like annotations) to read the corpus successfully are present. If files are
        missing, a list with the paths of the missing files is returned. All paths are relative to `path`. If no files
        are missing, an empty list is returned.

        Args:
            path (str): Path to the root directory of the data set

        Returns:
            list: Paths of all the missing files, relative to the path of the root directory of the data set.
        """
        return []


class CorpusWriter(metaclass=abc.ABCMeta):
    """
    Abstract class for writing a corpus.

    To implement a writer for a custom format, programmers are expected to subclass this class and to implement all
    abstract methods. The documentation of each abstract methods details the requirements that have to be met by an
    implementation.
    """

    def save(self, corpus, path):
        """
        Save the dataset at the given path.

        Args:
            corpus (Corpus): The corpus to save.
            path (str): Path to save the corpus to.
        """
        self._save(corpus, path)

    @classmethod
    @abc.abstractmethod
    def type(cls):
        """
        Returns a string that uniquely identifies the reader. This is usually the name of the corpus, for example
        `musan` or `timit`. Users can use this string to obtain an instance of the desired reader through
        :meth:`~audiomate.corpus.io.create_writer_of_type` or get a list of all built-in readers with
        :meth:`~audiomate.corpus.io.available_writers`.

        Returns:
            str: Name of the writer
        """
        return 'not_implemented'

    @abc.abstractmethod
    def _save(self, corpus, path):
        """
        Writes the corpus to disk to the given path.

        Args:
            corpus (Corpus): Corpus to write to disk
            path (str): Path of the target directory
        """
        pass



import os

import pytest
import requests_mock

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus.io import free_spoken_digits

from tests import resources
from . import reader_test as rt


@pytest.fixture()
def zip_data():
    path = resources.get_resource_path([
        'sample_files',
        'zip_sample_with_subfolder.zip'
    ])

    with open(path, 'rb') as f:
        return f.read()


class TestFreeSpokenDigitDownloader:

    def test_download(self, zip_data, tmpdir):
        target_folder = tmpdir.strpath
        downloader = free_spoken_digits.FreeSpokenDigitDownloader()

        with requests_mock.Mocker() as mock:
            mock.get(free_spoken_digits.MASTER_DOWNLOAD_URL, content=zip_data)

            downloader.download(target_folder)

        assert os.path.isfile(os.path.join(target_folder, 'a.txt'))
        assert os.path.isfile(os.path.join(target_folder, 'subsub', 'b.txt'))
        assert os.path.isfile(os.path.join(target_folder, 'subsub', 'c.txt'))


class TestFreeSpokenDigitReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('free_spoken_digits')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'recordings')

    EXPECTED_NUMBER_OF_TRACKS = 4
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('0_jackson_0', '0_jackson_0.wav'),
        rt.ExpFileTrack('1_jackson_0', '1_jackson_0.wav'),
        rt.ExpFileTrack('2_theo_0', '2_theo_0.wav'),
        rt.ExpFileTrack('2_theo_1', '2_theo_1.wav'),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 2
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('jackson', 2, issuers.Gender.UNKNOWN,
                      issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('theo', 2, issuers.Gender.UNKNOWN,
                      issuers.AgeGroup.UNKNOWN, None)
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 4
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('0_jackson_0', '0_jackson_0', 'jackson', 0, float('inf')),
        rt.ExpUtterance('1_jackson_0', '1_jackson_0', 'jackson', 0, float('inf')),
        rt.ExpUtterance('2_theo_0', '2_theo_0', 'theo', 0, float('inf')),
        rt.ExpUtterance('2_theo_1', '2_theo_1', 'theo', 0, float('inf'))
    ]

    EXPECTED_LABEL_LISTS = {
        '0_jackson_0': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        '1_jackson_0': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        '2_theo_0': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        '2_theo_1': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)
        ],
    }

    EXPECTED_LABELS = {
        '0_jackson_0': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, '0', 0, float('inf')),
        ],
        '1_jackson_0': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, '1', 0, float('inf')),
        ],
        '2_theo_0': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, '2', 0, float('inf')),
        ],
        '2_theo_1': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, '2', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 0

    def load(self):
        reader = free_spoken_digits.FreeSpokenDigitReader()
        return reader.load(self.SAMPLE_PATH)



from audiomate import corpus
from . import base


class UtteranceTranscriptionRatioValidator(base.Validator):
    """
    Checks if the ratio between utterance-duration and transcription-length is below a given ratio.
    This is used to find utterances where the speech transcription is to long for a given utterance,
    meaning too much characters per second.

    Args:
        max_characters_per_second (int): If char/sec of an utterance is higher than this it is returned.
        label_list_idx (str): The label-list to use for validation.
    """

    def __init__(self, max_characters_per_second=10, label_list_idx=corpus.LL_WORD_TRANSCRIPT):
        self.max_characters_per_second = max_characters_per_second
        self.label_list_idx = label_list_idx

    def name(self):
        return 'Utterance-Transcription-Ratio ({})'.format(self.label_list_idx)

    def validate(self, corpus):
        """
        Perform the validation on the given corpus.

        Args:
            corpus (Corpus): The corpus to test/validate.

        Returns:
            InvalidItemsResult: Validation result.
        """
        invalid_utterances = {}

        for utterance in corpus.utterances.values():
            duration = utterance.duration
            ll = utterance.label_lists[self.label_list_idx]

            # We count the characters of all labels
            transcription = ' '.join([l.value for l in ll])
            num_chars = len(transcription.replace(' ', ''))

            char_per_sec = num_chars / duration

            if char_per_sec > self.max_characters_per_second:
                invalid_utterances[utterance.idx] = char_per_sec

        passed = len(invalid_utterances) <= 0
        info = {
            'Threshold max. characters per second': str(self.max_characters_per_second),
            'Label-List ID': self.label_list_idx
        }

        return base.InvalidItemsResult(passed, invalid_utterances, name=self.name(), info=info)


class LabelCountValidator(base.Validator):
    """
    Checks if every utterance contains a label-list with the given id and has at least `min_number_of_labels`.

    Args:
        min_number_of_labels (int): Minimum number of expected labels.
        label_list_idx (str): The label-list to use for validation.
    """

    def __init__(self, min_number_of_labels=1, label_list_idx=corpus.LL_WORD_TRANSCRIPT):
        self.min_number_of_labels = min_number_of_labels
        self.label_list_idx = label_list_idx

    def name(self):
        return 'Label-Count ({})'.format(self.label_list_idx)

    def validate(self, corpus):
        """
        Perform the validation on the given corpus.

        Args:
            corpus (Corpus): The corpus to test/validate.

        Returns:
            InvalidItemsResult: Validation result.
        """
        invalid_utterances = {}

        for utterance in corpus.utterances.values():
            if self.label_list_idx in utterance.label_lists.keys():
                ll = utterance.label_lists[self.label_list_idx]

                if len(ll) < self.min_number_of_labels:
                    invalid_utterances[utterance.idx] = 'Only {} labels'.format(len(ll))
            else:
                invalid_utterances[utterance.idx] = 'No label-list {}'.format(self.label_list_idx)

        passed = len(invalid_utterances) <= 0
        info = {
            'Min. number of labels': str(self.min_number_of_labels),
            'Label-List ID': self.label_list_idx
        }

        return base.InvalidItemsResult(passed, invalid_utterances, name=self.name(), info=info)


class LabelCoverageValidationResult(base.ValidationResult):
    """
    Result of a the :class:`LabelCoverageValidator`.

    Args:
        passed (bool): A boolean indicating, if the validation has passed (True) or failed (False).
        uncovered_segments (dict): A dictionary containing a list of uncovered segments for every utterance.
        name (str): The name of the validator, that produced the result.
        info (dict): Dictionary containing key/value string-pairs with detailed information of the validation.
                     For example id of the label-list that was validated.
    """

    def __init__(self, passed, uncovered_segments, name, info=None):
        super(LabelCoverageValidationResult, self).__init__(passed, name=name, info=info)

        self.uncovered_segments = uncovered_segments

    def get_report(self):
        """
        Return a string containing a report of the result.
        This can used to print or save to a text file.

        Returns:
            str: String containing infos about the result
        """

        lines = [super(LabelCoverageValidationResult, self).get_report()]

        if len(self.uncovered_segments) > 0:
            lines.append('\nUncovered segments:')

            for utt_idx, utt_segments in self.uncovered_segments.items():
                if len(utt_segments) > 0:
                    lines.append('\n{}'.format(utt_idx))
                    sorted_items = sorted(utt_segments, key=lambda x: x[0])
                    lines.extend(['    * {:10.2f}  -  {:10.2f}'.format(x[0], x[1]) for x in sorted_items])

        return '\n'.join(lines)


class LabelCoverageValidator(base.Validator):
    """
    Check if every portion of the utterance is covered with at least one label.
    The validator returns segments (start, end) of an utterance, where no label is defined
    within the given label-list.

    Args:
        label_list_idx (str): The idx of the label-list to check.
        threshold (float): A threshold for the length of a segment to be considered as uncovered.
    """

    def __init__(self, label_list_idx, threshold=0.01):
        self.label_list_idx = label_list_idx
        self.threshold = threshold

    def name(self):
        return 'Label-Coverage ({})'.format(self.label_list_idx)

    def validate(self, corpus):
        """
        Perform the validation on the given corpus.

        Args:
            corpus (Corpus): The corpus to test/validate.

        Returns:
            LabelCoverageValidationResult: Validation result.
        """

        uncovered_segments = {}

        for utterance in corpus.utterances.values():
            utt_segments = self.validate_utterance(utterance)

            if len(utt_segments) > 0:
                uncovered_segments[utterance.idx] = utt_segments

        passed = len(uncovered_segments) <= 0
        info = {
            'Label-List ID': self.label_list_idx,
            'Threshold': str(self.threshold)
        }

        return LabelCoverageValidationResult(passed, uncovered_segments, self.name(), info)

    def validate_utterance(self, utterance):
        """
        Validate the given utterance and return a list of uncovered segments (start, end).
        """
        uncovered_segments = []

        if self.label_list_idx in utterance.label_lists.keys():
            start = 0
            end = utterance.duration
            ll = utterance.label_lists[self.label_list_idx]
            ranges = list(ll.ranges(yield_ranges_without_labels=True))

            # Check coverage at start
            if ranges[0][0] - start > self.threshold:
                uncovered_segments.append((start, ranges[0][0]))

            # Check for empty ranges
            for range in ranges:
                if len(range[2]) == 0 and range[1] - range[0] > self.threshold:
                    uncovered_segments.append((range[0], range[1]))

            # Check coverage at end
            if ranges[-1][1] > 0 and end - ranges[-1][1] > self.threshold:
                uncovered_segments.append((ranges[-1][1], end))

        else:
            uncovered_segments.append((utterance.start, utterance.end))

        return uncovered_segments


class LabelOverflowValidationResult(base.ValidationResult):
    """
    Result of a the :class:`LabelOverflowValidator`.

    Args:
        passed (bool): A boolean indicating, if the validation has passed (True) or failed (False).
        overflow_segments (dict): A dictionary containing a list of overflowing segments for every utterance.
        name (str): The name of the validator, that produced the result.
        info (dict): Dictionary containing key/value string-pairs with detailed information of the validation.
                     For example id of the label-list that was validated.
    """

    def __init__(self, passed, overflow_segments, name, info=None):
        super(LabelOverflowValidationResult, self).__init__(passed, name=name, info=info)

        self.overflow_segments = overflow_segments

    def get_report(self):
        """
        Return a string containing a report of the result.
        This can used to print or save to a text file.

        Returns:
            str: String containing infos about the result
        """

        lines = [super(LabelOverflowValidationResult, self).get_report()]

        if len(self.overflow_segments) > 0:
            lines.append('\nSegments outside of the utterance:')

            for utt_idx, utt_segments in self.overflow_segments.items():
                if len(utt_segments) > 0:
                    lines.append('\n{}'.format(utt_idx))
                    sorted_items = sorted(utt_segments, key=lambda x: x[0])
                    lines.extend(['    * {:10.2f}  -  {:10.2f} :  {}'.format(x[0], x[1], x[2]) for x in sorted_items])

        return '\n'.join(lines)


class LabelOverflowValidator(base.Validator):
    """
    Check if all labels are within the boundaries of an utterance.
    Finds all segments of labels that lie outside of an utterance.

    Args:
        label_list_idx (str): The idx of the label-list to check.
        threshold (float): A threshold for a time distance to be considered for an overflow.
    """

    def __init__(self, label_list_idx, threshold=0.01):
        self.label_list_idx = label_list_idx
        self.threshold = threshold

    def name(self):
        return 'Label-Overflow ({})'.format(self.label_list_idx)

    def validate(self, corpus):
        """
        Perform the validation on the given corpus.

        Args:
            corpus (Corpus): The corpus to test/validate.

        Returns:
            LabelOverflowValidationResult: Validation result.
        """

        overflow_segments = {}

        for utterance in corpus.utterances.values():
            utt_segments = self.validate_utterance(utterance)

            if len(utt_segments) > 0:
                overflow_segments[utterance.idx] = utt_segments

        passed = len(overflow_segments) <= 0
        info = {
            'Label-List ID': self.label_list_idx,
            'Threshold': str(self.threshold)
        }

        return LabelOverflowValidationResult(passed, overflow_segments, self.name(), info)

    def validate_utterance(self, utterance):
        """
        Validate the given utterance and return a list of segments (start, end, label-value),
        that are outside of the utterance.
        """
        overflow_segments = []

        if self.label_list_idx in utterance.label_lists.keys():
            ll = utterance.label_lists[self.label_list_idx]
            start = 0
            end = utterance.duration

            for label in ll:
                if start - label.start > self.threshold:
                    label_end = label.end if label.end != float('inf') else end
                    overflow_end = min(start, label_end)
                    overflow_segments.append((label.start, overflow_end, label.value))

                if label.end != float('inf') and label.end - end > self.threshold:
                    overflow_start = max(end, label.start)
                    overflow_segments.append((overflow_start, label.end, label.value))

        return overflow_segments



import os
import json

import audiomate
from audiomate import annotations
from audiomate.formats import audacity
from audiomate.utils import textfile
from . import base
from . import default

FILES_FILE_NAME = 'files.txt'
ISSUER_FILE_NAME = 'issuers.json'
UTTERANCE_FILE_NAME = 'utterances.txt'
UTT_ISSUER_FILE_NAME = 'utt_issuers.txt'
LABEL_FILE = 'labels.txt'
FEAT_CONTAINER_FILE_NAME = 'features.txt'


def extract_meta_from_label_value(label):
    meta_match = default.META_PATTERN.match(label.value)

    if meta_match is not None:
        meta_json = meta_match.group(2)
        label.meta = json.loads(meta_json)
        label.value = meta_match.group(1)


class BroadcastReader(base.CorpusReader):
    """
    Reads corpora in the Broadcast format.
    """

    @classmethod
    def type(cls):
        return 'broadcast'

    def _check_for_missing_files(self, path):
        necessary_files = [FILES_FILE_NAME, UTTERANCE_FILE_NAME]
        missing_files = []

        for file_name in necessary_files:
            file_path = os.path.join(path, file_name)

            if not os.path.isfile(file_path):
                missing_files.append(file_name)

        return missing_files

    def _load(self, path):
        file_path = os.path.join(path, FILES_FILE_NAME)
        issuer_path = os.path.join(path, ISSUER_FILE_NAME)
        utt_issuer_path = os.path.join(path, UTT_ISSUER_FILE_NAME)
        utterance_path = os.path.join(path, UTTERANCE_FILE_NAME)
        feat_path = os.path.join(path, FEAT_CONTAINER_FILE_NAME)

        corpus = audiomate.Corpus(path=path)

        default.DefaultReader.read_files(file_path, corpus)
        default.DefaultReader.read_issuers(issuer_path, corpus)
        utt_id_to_issuer = default.DefaultReader.read_utt_to_issuer_mapping(utt_issuer_path, corpus)
        default.DefaultReader.read_utterances(utterance_path, corpus, utt_id_to_issuer)
        BroadcastReader.read_labels(path, corpus)
        default.DefaultReader.read_feature_containers(feat_path, corpus)

        return corpus

    @staticmethod
    def read_labels(path, corpus):
        label_reference_file = os.path.join(path, LABEL_FILE)
        label_references = textfile.read_separated_lines(label_reference_file, separator=' ', max_columns=3)

        for record in label_references:
            utt_idx = record[0]
            label_path = os.path.join(path, record[1])
            label_idx = None

            if len(record) > 2:
                label_idx = record[2]

            ll = annotations.LabelList(idx=label_idx)

            for label in audacity.read_label_file(label_path):
                start = label[0]
                end = label[1]
                value = label[2]

                if end < 0:
                    end = float('inf')

                ll.addl(value, start, end)

            ll.apply(extract_meta_from_label_value)
            corpus.utterances[utt_idx].set_label_list(ll)



import os

import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.utils import textfile
from . import base
from . import downloader

DOWNLOAD_URL = 'http://www.openslr.org/resources/17/musan.tar.gz'

AUDIO_TYPES_ = ['music', 'noise', 'speech']

ANN_NUM_COLUMS_ = {'music': 4, 'noise': 1, 'speech': 3}

ANN_FILE_NAME_ = 'ANNOTATIONS'


class MusanDownloader(downloader.ArchiveDownloader):
    """
    Downloader for the MUSAN Corpus.

    Args:
        url (str): The url to download the dataset from. If not given the default URL is used.
                   It is expected to be a tar.gz file.
    """

    def __init__(self, url=None):
        if url is None:
            url = DOWNLOAD_URL

        super(MusanDownloader, self).__init__(
            url,
            move_files_up=True
        )

    @classmethod
    def type(cls):
        return 'musan'


class MusanReader(base.CorpusReader):
    """
    Reader for the MUSAN corpus. MUSAN is a corpus of music, speech, and noise recordings.

    .. seealso::

       `MUSAN: A Music, Speech, and Noise Corpus <https://arxiv.org/pdf/1510.08484v1.pdf>`_
          Paper explaining the structure and characteristics of the corpus

       `OpenSLR: MUSAN <http://www.openslr.org/17/>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'musan'

    def _check_for_missing_files(self, path):
        # Some label files are missing anyway in the original data set
        # (e.g. speech/us-gov/ANNOTATIONS). What's left would be checking for missing directories.
        return []

    def _load(self, path):
        create_or_get_issuer = {
            'music': self._create_or_get_music_issuer,
            'noise': self._create_or_get_noise_issuer,
            'speech': self._create_or_get_speech_issuer,
        }

        corpus = audiomate.Corpus(path=path)

        for type_name, type_directory in self._directories(path).items():
            for _, source_directory in self._directories(type_directory).items():
                labels_path = os.path.join(source_directory, ANN_FILE_NAME_)
                labels = {}

                if os.path.exists(labels_path):
                    labels = textfile.read_separated_lines_with_first_key(
                        labels_path, separator=' ', max_columns=ANN_NUM_COLUMS_[type_name])

                it = os.scandir(source_directory)

                for entry in it:
                    if not entry.name.endswith('.wav'):
                        continue

                    file_path = os.path.join(source_directory, entry.name)
                    file_idx = entry.name[0:-4]  # chop of .wav
                    utterance_idx = file_idx  # every file is a separate utterance
                    issuer_idx = create_or_get_issuer[type_name](corpus, file_idx, labels)

                    corpus.new_file(file_path, track_idx=file_idx, copy_file=False)
                    utterance = corpus.new_utterance(utterance_idx, file_idx, issuer_idx)
                    utterance.set_label_list(annotations.LabelList.create_single(
                        type_name, idx=audiomate.corpus.LL_DOMAIN))

        return corpus

    @staticmethod
    def _directories(path):
        directories = {}
        it = os.scandir(path)
        for entry in it:
            if not entry.is_dir():
                continue

            directories[entry.name] = os.path.join(path, entry.name)

        return directories

    # noinspection PyUnusedLocal
    @staticmethod
    def _create_or_get_noise_issuer(corpus, file_idx, labels):
        return None

    @staticmethod
    def _create_or_get_music_issuer(corpus, file_idx, labels):
        if file_idx not in labels:
            return None

        issuer_idx = labels[file_idx][2]

        if issuer_idx not in corpus.issuers:
            issuer = issuers.Artist(issuer_idx, name=issuer_idx)
            corpus.import_issuers(issuer)

        return issuer_idx

    @staticmethod
    def _create_or_get_speech_issuer(corpus, file_idx, labels):
        if file_idx not in labels:
            return None

        issuer = issuers.Speaker(file_idx)

        if file_idx in labels:
            if labels[file_idx][0] == 'm':
                issuer.gender = issuers.Gender.MALE
            elif labels[file_idx][0] == 'f':
                issuer.gender = issuers.Gender.FEMALE

        corpus.import_issuers(issuer)

        return file_idx



import numpy as np


class DataStats(object):
    """
    This class holds statistics for any kind of numerical data.

    Args:
        mean (float): The mean overall data points.
        var (float): The variance overall data points.
        min (float): The minimum value within all data points.
        max (float): The maximum value within all data points.
        num (int): The number of data points, these statistics were calculated of.
    """

    __slots__ = ['mean', 'var', 'min', 'max', 'num']

    def __init__(self, mean, var, min, max, num):
        self.mean = mean
        self.var = var
        self.min = min
        self.max = max
        self.num = num

    @property
    def values(self):
        """
        Return all values as numpy-array (mean, var, min, max, num).
        """
        return np.array([self.mean, self.var, self.min, self.max, self.num])

    def to_dict(self):
        """
        Return the stats as a dictionary.
        """
        return {
            'mean': self.mean,
            'var': self.var,
            'min': self.min,
            'max': self.max,
            'num': self.num
        }

    @classmethod
    def from_dict(self, dict_with_stats):
        """
        Create a DataStats object from a dictionary with stats.

        Args:
            dict_with_stats (dict): Dictionary containing stats.

        Returns:
            (DataStats): Statistics
        """

        return DataStats(dict_with_stats['mean'],
                         dict_with_stats['var'],
                         dict_with_stats['min'],
                         dict_with_stats['max'],
                         dict_with_stats['num'])

    @classmethod
    def concatenate(cls, list_of_stats):
        """
        Take a list of stats from different sets of data points and
        merge the stats for getting stats overall data points.

        Args:
            list_of_stats (iterable): A list containing stats for different sets of data points.

        Returns:
            DataStats: Stats calculated overall sets of data points.
        """

        all_stats = np.stack([stats.values for stats in list_of_stats])
        all_counts = all_stats[:, 4]
        all_counts_relative = all_counts / np.sum(all_counts)

        min_value = float(np.min(all_stats[:, 2]))
        max_value = float(np.max(all_stats[:, 3]))
        mean_value = float(np.sum(all_counts_relative * all_stats[:, 0]))
        var_value = float(np.sum(all_counts_relative * (all_stats[:, 1] + np.power(all_stats[:, 0] - mean_value, 2))))
        num_value = int(np.sum(all_counts))

        return cls(mean_value, var_value, min_value, max_value, num_value)



import os

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus.io import fluent_speech

from tests import resources
from . import reader_test as rt


class TestFreeSpokenDigitReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('fluent_speech')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'wavs', 'speakers')

    EXPECTED_NUMBER_OF_TRACKS = 8
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('x', 'a/x.wav'),
        rt.ExpFileTrack('y', 'a/y.wav'),
        rt.ExpFileTrack('z', 'b/z.wav'),
        rt.ExpFileTrack('p', 'g/p.wav'),
        rt.ExpFileTrack('m', 'g/m.wav'),
        rt.ExpFileTrack('i', 'e/i.wav'),
        rt.ExpFileTrack('o', 'f/o.wav'),
        rt.ExpFileTrack('k', 'f/k.wav'),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 5
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('a', 2, issuers.Gender.MALE, issuers.AgeGroup.ADULT, 'eng'),
        rt.ExpSpeaker('b', 1, issuers.Gender.MALE, issuers.AgeGroup.SENIOR, 'eng'),
        rt.ExpSpeaker('e', 1, issuers.Gender.MALE, issuers.AgeGroup.ADULT, 'eng'),
        rt.ExpSpeaker('f', 2, issuers.Gender.MALE, issuers.AgeGroup.ADULT, 'fra'),
        rt.ExpSpeaker('g', 2, issuers.Gender.FEMALE, issuers.AgeGroup.ADULT, 'tel'),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 8
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('x', 'x', 'a', 0, float('inf')),
        rt.ExpUtterance('y', 'y', 'a', 0, float('inf')),
        rt.ExpUtterance('z', 'z', 'b', 0, float('inf')),
        rt.ExpUtterance('p', 'p', 'g', 0, float('inf')),
        rt.ExpUtterance('m', 'p', 'g', 0, float('inf')),
        rt.ExpUtterance('i', 'i', 'e', 0, float('inf')),
        rt.ExpUtterance('o', 'o', 'f', 0, float('inf')),
        rt.ExpUtterance('k', 'k', 'f', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        'i': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList('action', 1),
            rt.ExpLabelList('location', 1),
        ],
        'o': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList('action', 1),
            rt.ExpLabelList('object', 1),
            rt.ExpLabelList('location', 1),
        ],
        'k': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList('action', 1),
            rt.ExpLabelList('object', 1),
        ],
        'x': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList('action', 1),
        ],
        'y': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList('action', 1),
            rt.ExpLabelList('object', 1),
        ],
        'z': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList('action', 1),
            rt.ExpLabelList('object', 1),
        ],
        'p': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList('action', 1),
            rt.ExpLabelList('object', 1),
            rt.ExpLabelList('location', 1),
        ],
        'm': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList('action', 1),
            rt.ExpLabelList('object', 1),
        ],
    }

    EXPECTED_LABELS = {
        'i': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Change', 0, float('inf')),
            rt.ExpLabel('action', 'change', 0, float('inf')),
            rt.ExpLabel('location', 'kitchen', 0, float('inf')),
        ],
        'o': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Resume', 0, float('inf')),
            rt.ExpLabel('action', 'activate', 0, float('inf')),
            rt.ExpLabel('object', 'music', 0, float('inf')),
            rt.ExpLabel('location', 'kitchen', 0, float('inf')),
        ],
        'k': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Turn off the lights', 0, float('inf')),
            rt.ExpLabel('action', 'activate', 0, float('inf')),
            rt.ExpLabel('object', 'lights', 0, float('inf')),
        ],
        'x': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Change language', 0, float('inf')),
            rt.ExpLabel('action', 'change language', 0, float('inf')),
        ],
        'y': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Resume', 0, float('inf')),
            rt.ExpLabel('action', 'activate', 0, float('inf')),
            rt.ExpLabel('object', 'music', 0, float('inf')),
        ],
        'z': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Turn the lights on', 0, float('inf')),
            rt.ExpLabel('action', 'activate', 0, float('inf')),
            rt.ExpLabel('object', 'lights', 0, float('inf')),
        ],
        'p': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'lights', 0, float('inf')),
            rt.ExpLabel('action', 'change', 0, float('inf')),
            rt.ExpLabel('object', 'heat', 0, float('inf')),
            rt.ExpLabel('location', 'bedroom', 0, float('inf')),
        ],
        'm': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'lamp', 0, float('inf')),
            rt.ExpLabel('action', 'activate', 0, float('inf')),
            rt.ExpLabel('object', 'music', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 3
    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('test', [
            'i',
            'o',
            'k',
        ]),
        rt.ExpSubview('train', [
            'x',
            'y',
            'z',
        ]),
        rt.ExpSubview('valid', [
            'p',
            'm',
        ]),
    ]

    def load(self):
        reader = fluent_speech.FluentSpeechReader()
        return reader.load(self.SAMPLE_PATH)



"""
This module contains any functionality for downloading and extracting data from any remotes.
"""

import zipfile
import tarfile
import requests


def download_file(url, target_path):
    """
    Download the file from the given `url` and store it at `target_path`.
    """

    r = requests.get(url, stream=True)

    with open(target_path, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)


def extract_zip(zip_path, target_folder):
    """
    Extract the content of the zip-file at `zip_path` into `target_folder`.
    """
    with zipfile.ZipFile(zip_path) as archive:
        archive.extractall(target_folder)


def extract_tar(tar_path, target_folder):
    """
    Extract the content of the tar-file at `tar_path` into `target_folder`.
    """
    with tarfile.open(tar_path, 'r') as archive:
        archive.extractall(target_folder)



import librosa
import numpy as np

from . import base
from . import spectral


class Tempogram(base.Computation):
    """
    Computation step to compute tempogram

    Based on http://librosa.github.io/librosa/generated/librosa.feature.tempogram.html

    Args:
        n_mels (int): Number of mel bands to generate.
        win_length (int): Length of the onset autocorrelation window (in frames/onset measurements).
                          The default settings (384) corresponds to 384 * hop_length / sr ~= 8.9s.
    """

    def __init__(self, n_mels=128, win_length=384, parent=None, name=None):
        super(Tempogram, self).__init__(min_frames=win_length, left_context=1, right_context=0,
                                        parent=parent, name=name)

        self.n_mels = n_mels
        self.win_length = win_length

        self.rest = None

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        # Cleanup rest if it's the first frame
        if chunk.offset == 0:
            self.rest = None

        # Compute mel-spectrogram
        power_spec = np.abs(spectral.stft_from_frames(chunk.data.T)) ** 2
        mel = np.abs(librosa.feature.melspectrogram(S=power_spec, n_mels=self.n_mels, sr=sampling_rate))
        mel_power = librosa.power_to_db(mel)

        # Compute onset strengths
        oenv = librosa.onset.onset_strength(S=mel_power, center=False)

        # Remove context, otherwise we have duplicate frames while online processing
        oenv = oenv[chunk.left_context:]

        if self.rest is not None:
            all_frames = np.concatenate([self.rest, oenv])
        else:
            # Its the first chunk --> pad to center tempogram windows at the beginning
            all_frames = np.pad(oenv, (self.win_length // 2, 0), mode='linear_ramp', end_values=0)

        if chunk.is_last:
            # Its the last chunk --> pad to center tempogram windows at end
            all_frames = np.pad(all_frames, (0, self.win_length // 2), mode='linear_ramp', end_values=0)

            # Compensate the 1 frame that is too much since we want win-len - 1 additional frames,
            # With an even win-len we would have win-len additional frames
            if self.win_length % 2 == 0:
                all_frames = all_frames[:-1]

        if all_frames.shape[0] >= self.win_length:
            tempogram = librosa.feature.tempogram(onset_envelope=all_frames, sr=sampling_rate,
                                                  win_length=self.win_length, center=False).T

            self.rest = all_frames[tempogram.shape[0]:]

            return tempogram
        else:
            self.rest = all_frames



import pytest

import numpy as np

from audiomate.utils import stats


class TestDataStats:

    def test_values(self):
        s = stats.DataStats(2.3, 1.2, 4.0, -2, 99)
        assert np.array_equal(s.values, np.array([2.3, 1.2, 4.0, -2, 99]))

    def test_concatenate(self):
        values = np.random.randn(20).reshape(4, 5)

        stats_list = []

        for x in values:
            s = stats.DataStats(
                float(x.mean()),
                float(x.var()),
                x.min(),
                x.max(),
                x.size
            )
            stats_list.append(s)

        concatenated = stats.DataStats.concatenate(stats_list)

        assert concatenated.mean == pytest.approx(np.mean(values))
        assert concatenated.var == pytest.approx(np.var(values))
        assert concatenated.min == pytest.approx(np.min(values))
        assert concatenated.max == pytest.approx(np.max(values))
        assert concatenated.num == values.size

    def test_to_dict(self):
        s = stats.DataStats(2.3, 1.2, -2, 4.0, 99)
        d = s.to_dict()

        assert d == {'mean': 2.3, 'var': 1.2, 'max': 4.0, 'min': -2, 'num': 99}

    def test_from_dict(self):
        s = stats.DataStats.from_dict({
            'mean': 2.3,
            'var': 1.2,
            'max': 4.0,
            'min': -2,
            'num': 99
        })

        assert s.mean == pytest.approx(2.3)
        assert s.var == pytest.approx(1.2)
        assert s.min == pytest.approx(-2)
        assert s.max == pytest.approx(4.0)
        assert s.num == 99



import os.path

from audiomate import annotations
from audiomate.formats import audacity


class TestAudacityFormat:

    def test_read_label_file_en(self):
        path = os.path.join(os.path.dirname(__file__), 'audacity_labels_en.txt')
        labels = audacity.read_label_file(path)

        assert len(labels) == 2

        assert labels[0][0] == 43352.824046
        assert labels[0][1] == 43525.837661
        assert labels[0][2] == 'music'

        assert labels[1][0] == 43512.446969
        assert labels[1][1] == 43531.343483
        assert labels[1][2] == 'speech_male'

    def test_read_label_file_de(self):
        path = os.path.join(os.path.dirname(__file__), 'audacity_labels_de.txt')
        labels = audacity.read_label_file(path)

        assert len(labels) == 2

        assert labels[0][0] == 43352.824046
        assert labels[0][1] == 43525.837661
        assert labels[0][2] == 'music'

        assert labels[1][0] == 43512.446969
        assert labels[1][1] == 43531.343483
        assert labels[1][2] == 'speech_male'

    def test_read_label_file_with_empty_value(self):
        path = os.path.join(os.path.dirname(__file__), 'audacity_labels_empty_value.txt')
        labels = audacity.read_label_file(path)

        assert len(labels) == 3

        assert labels[0][0] == 1
        assert labels[0][1] == 4
        assert labels[0][2] == 'music'

        assert labels[1][0] == 4
        assert labels[1][1] == 7
        assert labels[1][2] == ''

        assert labels[2][0] == 7
        assert labels[2][1] == 9
        assert labels[2][2] == 'speech_male'

    def test_read_label_list_en(self):
        path = os.path.join(os.path.dirname(__file__), 'audacity_labels_en.txt')
        ll = audacity.read_label_list(path)

        assert ll == annotations.LabelList(labels=[
            annotations.Label('music', 43352.824046, 43525.837661),
            annotations.Label('speech_male', 43512.446969, 43531.343483),
        ])

    def test_read_label_list_de(self):
        path = os.path.join(os.path.dirname(__file__), 'audacity_labels_de.txt')
        ll = audacity.read_label_list(path)

        assert ll == annotations.LabelList(labels=[
            annotations.Label('music', 43352.824046, 43525.837661),
            annotations.Label('speech_male', 43512.446969, 43531.343483),
        ])

    def test_read_label_list_with_empty_value(self):
        path = os.path.join(os.path.dirname(__file__), 'audacity_labels_empty_value.txt')
        ll = audacity.read_label_list(path)

        assert ll == annotations.LabelList(labels=[
            annotations.Label('music', 1, 4),
            annotations.Label('', 4, 7),
            annotations.Label('speech_male', 7, 9),
        ])

    def test_write_label_file(self, tmpdir):
        path = os.path.join(tmpdir.strpath, 'audacity_labels.txt')
        entries = [
            [10.01, 11.08, 'music'],
            [11.08, 13.33, 'speech_male']
        ]

        audacity.write_label_file(path, entries)

        assert os.path.isfile(path)

        with open(path) as file:
            lines = file.readlines()

            assert len(lines) == 2

            assert lines[0] == '10.01\t11.08\tmusic\n'
            assert lines[1] == '11.08\t13.33\tspeech_male\n'

    def test_write_label_list(self, tmpdir):
        path = os.path.join(tmpdir.strpath, 'audacity_labels.txt')
        ll = annotations.LabelList(labels=[
            annotations.Label('music', 10.01, 11.08),
            annotations.Label('speech_male', 11.08, 13.33),
        ])

        audacity.write_label_list(path, ll)

        assert os.path.isfile(path)

        with open(path) as file:
            lines = file.readlines()

            assert len(lines) == 2

            assert lines[0] == '10.01\t11.08\tmusic\n'
            assert lines[1] == '11.08\t13.33\tspeech_male\n'



import os

from audiomate.utils import files


def test_move_all_files_from_subfolders_to_top(tmpdir):
    base_path = tmpdir.strpath

    os.makedirs(os.path.join(base_path, 'a', 'asub'))
    os.makedirs(os.path.join(base_path, 'b'))

    open(os.path.join(base_path, 'a', '1.txt'), 'w').close()
    open(os.path.join(base_path, 'a', 'asub', '3.txt'), 'w').close()
    open(os.path.join(base_path, 'b', '2.txt'), 'w').close()

    files.move_all_files_from_subfolders_to_top(base_path)

    assert len(os.listdir(base_path)) == 5

    assert os.path.isdir(os.path.join(base_path, 'a'))
    assert os.path.isdir(os.path.join(base_path, 'b'))
    assert os.path.isdir(os.path.join(base_path, 'asub'))
    assert os.path.isfile(os.path.join(base_path, '1.txt'))
    assert os.path.isfile(os.path.join(base_path, '2.txt'))


def test_move_all_files_from_subfolders_to_top_and_delete_subfolders(tmpdir):
    base_path = tmpdir.strpath

    os.makedirs(os.path.join(base_path, 'a', 'asub'))
    os.makedirs(os.path.join(base_path, 'b'))

    open(os.path.join(base_path, 'a', '1.txt'), 'w').close()
    open(os.path.join(base_path, 'a', 'asub', '3.txt'), 'w').close()
    open(os.path.join(base_path, 'b', '2.txt'), 'w').close()

    files.move_all_files_from_subfolders_to_top(base_path, delete_subfolders=True)

    assert len(os.listdir(base_path)) == 3

    assert os.path.isdir(os.path.join(base_path, 'asub'))
    assert os.path.isfile(os.path.join(base_path, '1.txt'))
    assert os.path.isfile(os.path.join(base_path, '2.txt'))



import pytest

import audiomate
from audiomate import annotations
from audiomate import tracks
from audiomate import issuers

from tests import resources


@pytest.fixture
def ds():
    return resources.create_multi_label_corpus()


class TestCorpusView:

    def test_contains_track_returns_true(self, ds):
        res = ds.contains_track(ds.tracks['wav-1'])
        assert res

    def test_contains_track_returns_false(self, ds):
        track = tracks.FileTrack('wav-1', '/some/other/path/here')
        res = ds.contains_track(track)
        assert not res

    def test_contains_issuer_returns_true(self, ds):
        res = ds.contains_issuer(ds.issuers['spk-1'])
        assert res

    def test_contains_issuer_returns_false(self, ds):
        issuer = issuers.Issuer('spk-1')
        res = ds.contains_issuer(issuer)
        assert not res

    def test_all_label_values(self, ds):
        assert ds.all_label_values() == {'music', 'speech'}

    def test_label_count(self, ds):
        assert ds.label_count() == {'music': 11, 'speech': 7}

    def test_stats(self):
        ds = audiomate.Corpus.load(resources.sample_corpus_path('default'), reader='default')
        stats = ds.stats()

        assert stats.min == pytest.approx(-1.0)
        assert stats.max == pytest.approx(0.99996948)
        assert stats.mean == pytest.approx(-0.00013355668)
        assert stats.var == pytest.approx(0.015060359)

    def test_stats_per_utterance(self):
        ds = audiomate.Corpus.load(resources.sample_corpus_path('default'), reader='default')
        stats = ds.stats_per_utterance()

        assert set(list(stats.keys())) == {'utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'}

        assert stats['utt-1'].min == pytest.approx(-1.0)
        assert stats['utt-1'].max == pytest.approx(0.99996948)
        assert stats['utt-1'].mean == pytest.approx(-0.00023601724)
        assert stats['utt-1'].var == pytest.approx(0.017326673)
        assert stats['utt-1'].num == 118240

        assert stats['utt-3'].min == pytest.approx(-0.92578125)
        assert stats['utt-3'].max == pytest.approx(0.99996948)
        assert stats['utt-3'].mean == pytest.approx(-0.00041901905)
        assert stats['utt-3'].var == pytest.approx(0.017659103)
        assert stats['utt-3'].num == 24000

    def test_label_duration(self, ds):
        durations = ds.label_durations()

        assert durations['music'] == pytest.approx(44.0)
        assert durations['speech'] == pytest.approx(45.0)

    def test_duration(self, ds):
        duration = ds.total_duration

        assert duration == pytest.approx(85.190375)

    def test_all_tokens(self):
        corpus = resources.create_dataset()
        assert corpus.all_tokens() == {'who', 'am', 'i', 'are', 'is', 'he', 'you', 'she', 'they'}

    def test_all_tokens_returns_only_from_selected_label_lists(self):
        corpus = resources.create_dataset()
        ll = annotations.LabelList(idx='test', labels=[
            annotations.Label('what can he do')
        ])
        corpus.utterances['utt-1'].set_label_list(ll)

        target_lls = [audiomate.corpus.LL_WORD_TRANSCRIPT]
        expected_tokens = {'who', 'am', 'i', 'are', 'is', 'he', 'you', 'she', 'they'}
        assert corpus.all_tokens(label_list_ids=target_lls) == expected_tokens

    def test_all_tokens_with_custom_delimiter(self):
        corpus = resources.create_dataset()
        ll = annotations.LabelList(idx='test', labels=[
            annotations.Label('a, b, a, c')
        ])
        corpus.utterances['utt-1'].set_label_list(ll)

        target_lls = ['test']
        expected_tokens = {'a', 'b', 'c'}
        assert corpus.all_tokens(delimiter=',', label_list_ids=target_lls) == expected_tokens

    def test_split_utterances_to_max_time(self):
        corpus = resources.create_dataset()
        result = corpus.split_utterances_to_max_time(max_time=1.5, overlap=0.5)

        assert result.num_tracks == corpus.num_tracks
        assert result.num_issuers == corpus.num_issuers
        assert result.num_utterances == 8

        utt_len = 2.5951875

        assert result.utterances['utt-1_0'].start == 0.0
        assert result.utterances['utt-1_0'].end == utt_len / 2.0 + 0.5
        assert result.utterances['utt-1_1'].start == utt_len / 2.0 - 0.5
        assert result.utterances['utt-1_1'].end == float('inf')

        assert result.utterances['utt-2_0'].start == 0.0
        assert result.utterances['utt-2_0'].end == utt_len / 2.0 + 0.5
        assert result.utterances['utt-2_1'].start == utt_len / 2.0 - 0.5
        assert result.utterances['utt-2_1'].end == float('inf')

        assert result.utterances['utt-3'].start == 0.0
        assert result.utterances['utt-3'].end == 1.5

        assert result.utterances['utt-4'].start == 1.5
        assert result.utterances['utt-4'].end == 2.5

        assert result.utterances['utt-5_0'].start == 0.0
        assert result.utterances['utt-5_0'].end == utt_len / 2.0 + 0.5
        assert result.utterances['utt-5_1'].start == utt_len / 2.0 - 0.5
        assert result.utterances['utt-5_1'].end == float('inf')



import os

import numpy as np

from audiomate import corpus
from audiomate import tracks
from audiomate import issuers
from audiomate.corpus import io
from audiomate.utils import textfile

import pytest

from tests import resources
from . import reader_test as rt


@pytest.fixture
def writer():
    return io.KaldiWriter()


class TestKaldiReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('kaldi')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'files')

    EXPECTED_NUMBER_OF_TRACKS = 4
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('file-1', 'wav_1.wav'),
        rt.ExpFileTrack('file-2', 'wav_2.wav'),
        rt.ExpFileTrack('file-3', 'wav_3.wav'),
        rt.ExpFileTrack('file-4', 'wav_4.wav'),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 3
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('speaker-1', 2, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('speaker-2', 2, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('speaker-3', 1, issuers.Gender.FEMALE, issuers.AgeGroup.UNKNOWN, None),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 5
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('utt-1', 'file-1', 'speaker-1', 0, float('inf')),
        rt.ExpUtterance('utt-2', 'file-2', 'speaker-1', 0, float('inf')),
        rt.ExpUtterance('utt-3', 'file-3', 'speaker-2', 0, 15),
        rt.ExpUtterance('utt-4', 'file-3', 'speaker-2', 15, 25),
        rt.ExpUtterance('utt-5', 'file-4', 'speaker-3', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        'utt-1': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        'utt-2': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        'utt-3': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        'utt-4': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        'utt-5': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
    }

    EXPECTED_LABELS = {
        'utt-4': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'who are they', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 0

    def load(self):
        return io.KaldiReader().load(self.SAMPLE_PATH)


class TestKaldiWriter:

    def test_save(self, writer, tmpdir):
        ds = resources.create_dataset()
        path = tmpdir.strpath
        writer.save(ds, path)

        assert 'segments' in os.listdir(path)
        assert 'text' in os.listdir(path)
        assert 'utt2spk' in os.listdir(path)
        assert 'wav.scp' in os.listdir(path)

    def test_save_spk2gender(self, writer, tmpdir):
        writer = io.KaldiWriter(create_spk2gender=True)
        ds = resources.create_dataset()
        path = tmpdir.strpath
        writer.save(ds, path)

        assert 'spk2gender' in os.listdir(path)

        content = textfile.read_separated_lines(
            os.path.join(path, 'spk2gender'),
            separator=' ',
            max_columns=2
        )

        assert content[0][0] == 'spk-1'
        assert content[0][1] == 'm'

        assert content[1][0] == 'spk-2'
        assert content[1][1] == 'f'

        assert content[2][0] == 'spk-3'
        assert content[2][1] == 'm'

    def test_save_spk2gender_with_custom_default_gender(self, tmpdir):
        writer = io.KaldiWriter(create_spk2gender=True, default_gender='ui')
        ds = resources.create_dataset()
        path = tmpdir.strpath
        writer.save(ds, path)

        assert 'spk2gender' in os.listdir(path)

        content = textfile.read_separated_lines(
            os.path.join(path, 'spk2gender'),
            separator=' ',
            max_columns=2
        )

        assert content[0][0] == 'spk-1'
        assert content[0][1] == 'm'

        assert content[1][0] == 'spk-2'
        assert content[1][1] == 'f'

        assert content[2][0] == 'spk-3'
        assert content[2][1] == 'ui'

    def test_write_wav_scp(self, writer, tmpdir):
        wav_base = resources.get_resource_path(['wav_files'])
        wav_base = os.path.abspath(wav_base)

        ds = resources.create_dataset()
        ds.tracks['wav-1'].path = os.path.join(wav_base, 'wav_1.mp3')
        path = tmpdir.strpath
        writer.save(ds, path)

        content = textfile.read_separated_lines(
            os.path.join(path, 'wav.scp'),
            separator=' ',
            max_columns=2
        )

        assert content[0][0] == 'wav-1'
        assert content[0][1] == 'sox {} -t wav - |'.format(os.path.join(wav_base, 'wav_1.mp3'))
        assert content[1][0] == 'wav_2'
        assert content[1][1] == os.path.join(wav_base, 'wav_2.wav')
        assert content[2][0] == 'wav_3'
        assert content[2][1] == os.path.join(wav_base, 'wav_3.wav')
        assert content[3][0] == 'wav_4'
        assert content[3][1] == os.path.join(wav_base, 'wav_4.wav')

    def test_write_segments(self, writer, tmpdir):
        ds = resources.create_dataset()
        path = tmpdir.strpath
        writer.save(ds, path)

        content = textfile.read_separated_lines(
            os.path.join(path, 'segments'),
            separator=' ',
            max_columns=4
        )

        assert content[0][0] == 'spk-1-utt-1'
        assert content[0][1] == 'wav-1'
        assert float(content[0][2]) == 0
        assert float(content[0][3]) == -1

        assert content[1][0] == 'spk-1-utt-2'
        assert content[1][1] == 'wav_2'
        assert float(content[1][2]) == 0
        assert float(content[1][3]) == -1

        assert content[2][0] == 'spk-2-utt-3'
        assert content[2][1] == 'wav_3'
        assert float(content[2][2]) == pytest.approx(0)
        assert float(content[2][3]) == pytest.approx(1.5)

        assert content[3][0] == 'spk-2-utt-4'
        assert content[3][1] == 'wav_3'
        assert float(content[3][2]) == pytest.approx(1.5)
        assert float(content[3][3]) == pytest.approx(2.5)

        assert content[4][0] == 'spk-3-utt-5'
        assert content[4][1] == 'wav_4'
        assert float(content[4][2]) == 0
        assert float(content[4][3]) == -1

    def test_write_segments_absolute_times(self, writer, tmpdir):
        writer = io.KaldiWriter(use_absolute_times=True)
        ds = resources.create_dataset()
        path = tmpdir.strpath
        writer.save(ds, path)

        content = textfile.read_separated_lines(
            os.path.join(path, 'segments'),
            separator=' ',
            max_columns=4
        )

        assert content[0][0] == 'spk-1-utt-1'
        assert content[0][1] == 'wav-1'
        assert float(content[0][2]) == 0
        assert float(content[0][3]) == pytest.approx(2.5951875)

        assert content[1][0] == 'spk-1-utt-2'
        assert content[1][1] == 'wav_2'
        assert float(content[1][2]) == 0
        assert float(content[1][3]) == pytest.approx(2.5951875)

        assert content[2][0] == 'spk-2-utt-3'
        assert content[2][1] == 'wav_3'
        assert float(content[2][2]) == pytest.approx(0)
        assert float(content[2][3]) == pytest.approx(1.5)

        assert content[3][0] == 'spk-2-utt-4'
        assert content[3][1] == 'wav_3'
        assert float(content[3][2]) == pytest.approx(1.5)
        assert float(content[3][3]) == pytest.approx(2.5)

        assert content[4][0] == 'spk-3-utt-5'
        assert content[4][1] == 'wav_4'
        assert float(content[4][2]) == 0
        assert float(content[4][3]) == pytest.approx(2.5951875)

    def test_write_segments_no_speaker_prefix(self, tmpdir):
        writer = io.KaldiWriter(prefix_utterances_with_speaker=False)
        ds = resources.create_dataset()
        path = tmpdir.strpath
        writer.save(ds, path)

        content = textfile.read_separated_lines(
            os.path.join(path, 'segments'),
            separator=' ',
            max_columns=4
        )

        assert content[0][0] == 'utt-1'
        assert content[0][1] == 'wav-1'
        assert float(content[0][2]) == 0
        assert float(content[0][3]) == -1

        assert content[1][0] == 'utt-2'
        assert content[1][1] == 'wav_2'
        assert float(content[1][2]) == 0
        assert float(content[1][3]) == -1

        assert content[2][0] == 'utt-3'
        assert content[2][1] == 'wav_3'
        assert float(content[2][2]) == pytest.approx(0)
        assert float(content[2][3]) == pytest.approx(1.5)

        assert content[3][0] == 'utt-4'
        assert content[3][1] == 'wav_3'
        assert float(content[3][2]) == pytest.approx(1.5)
        assert float(content[3][3]) == pytest.approx(2.5)

        assert content[4][0] == 'utt-5'
        assert content[4][1] == 'wav_4'
        assert float(content[4][2]) == 0
        assert float(content[4][3]) == -1

    def test_write_utt2spk(self, writer, tmpdir):
        ds = resources.create_dataset()

        # Add utt without issuer
        # so in utt2spk it ends up with "utt-idx utt-idx"
        ds.new_file('/random/path', 'wav-33')
        ds.new_utterance('utt-23', 'wav-33')

        path = tmpdir.strpath
        writer.save(ds, path)

        content = textfile.read_separated_lines(
            os.path.join(path, 'utt2spk'),
            separator=' ',
            max_columns=2
        )

        assert content[0][0] == 'spk-1-utt-1'
        assert content[0][1] == 'spk-1'

        assert content[1][0] == 'spk-1-utt-2'
        assert content[1][1] == 'spk-1'

        assert content[2][0] == 'spk-2-utt-3'
        assert content[2][1] == 'spk-2'

        assert content[3][0] == 'spk-2-utt-4'
        assert content[3][1] == 'spk-2'

        assert content[4][0] == 'spk-3-utt-5'
        assert content[4][1] == 'spk-3'

        assert content[5][0] == 'utt-23'
        assert content[5][1] == 'utt-23'

    def test_exports_wavs_from_container_tracks(self, writer, tmpdir):
        path = tmpdir.strpath
        container_ds_path = os.path.join(path, 'container_ds')
        out_path = os.path.join(path, 'export')

        ds = resources.create_dataset()
        ds.relocate_audio_to_single_container(container_ds_path)

        writer.save(ds, out_path)

        print(os.listdir(out_path))

        track_path = os.path.join(out_path, 'audio', 'wav-1.wav')
        track = tracks.FileTrack(None, track_path)
        assert os.path.isfile(track_path)
        assert track.duration == pytest.approx(2.5951875)
        assert np.allclose(
            track.read_samples(),
            ds.tracks['wav-1'].read_samples(),
            atol=1e-05
        )

        track_path = os.path.join(out_path, 'audio', 'wav_2.wav')
        track = tracks.FileTrack(None, track_path)
        assert os.path.isfile(track_path)
        assert track.duration == pytest.approx(2.5951875)

        track_path = os.path.join(out_path, 'audio', 'wav_3.wav')
        track = tracks.FileTrack(None, track_path)
        assert os.path.isfile(track_path)
        assert track.duration == pytest.approx(2.5951875)

        track_path = os.path.join(out_path, 'audio', 'wav_4.wav')
        track = tracks.FileTrack(None, track_path)
        assert os.path.isfile(track_path)
        assert track.duration == pytest.approx(2.5951875)



import collections
import random

import numpy as np


def absolute_proportions(proportions, count):
    """
    Split a given integer into n parts according to len(proportions) so they sum up to count and
    match the given proportions.

    Args:
        proportions (dict): Dict of proportions, with a identifier as key.

    Returns:
        dict: Dictionary with absolute proportions and same identifiers as key.

    Example::

        >>> absolute_proportions({'train': 0.5, 'test': 0.5}, 100)
        {'train': 50, 'test': 50}
    """

    # first create absolute values by flooring non-integer portions
    relative_sum = sum(proportions.values())
    absolute_proportions = {idx: int(count / relative_sum * prop_value) for idx, prop_value in
                            proportions.items()}

    # Now distribute the rest value randomly over the different parts
    absolute_sum = sum(absolute_proportions.values())
    rest_value = count - absolute_sum
    subset_keys = sorted(list(proportions.keys()))

    for i in range(rest_value):
        key = subset_keys[i % len(subset_keys)]
        absolute_proportions[key] += 1

    return absolute_proportions


def split_identifiers(identifiers=[], proportions={}, seed=None):
    """
    Split the given identifiers by the given proportions.
    This function is deterministic, given the same seed.
    First the identifiers are sorted before shuffled using the given seed.

    Args:
        identifiers (list): List of identifiers (str).
        proportions (dict): A dictionary containing the proportions with the identifier from the
        input as key.
        seed (int): Seed to use for random operations.

    Returns:
        dict: Dictionary containing a list of identifiers per part with the same key as the
        proportions dict.

    Example::

        >>> split_identifiers(
        >>>     identifiers=['a', 'b', 'c', 'd'],
        >>>     proportions={'melvin' : 0.5, 'timmy' : 0.5}
        >>> )
        {'melvin' : ['a', 'c'], 'timmy' : ['b', 'd']}
    """

    identifiers = sorted(identifiers)

    # Shuffle to get random order
    rand = random.Random()
    rand.seed(a=seed)
    rand.shuffle(identifiers)

    abs_proportions = absolute_proportions(proportions, len(identifiers))

    parts = {}
    start_index = 0

    for idx in sorted(abs_proportions.keys()):
        proportion = abs_proportions[idx]
        parts[idx] = identifiers[start_index:start_index + proportion]
        start_index += proportion

    return parts


def get_identifiers_splitted_by_weights(identifiers={}, proportions={}, seed=None):
    """
    Divide the given identifiers based on the given proportions. But instead of randomly split
    the identifiers it is based on category weights. Every identifier has a weight for any
    number of categories. The target is, to split the identifiers in a way, so the sum of
    category k within part x is proportional to the sum of category x over all parts
    according to the given proportions. This is done by greedily insert the identifiers step by
    step in a part which has free space (weight). If there are no fitting parts anymore, the one
    with the least weight exceed is used.
    This function is deterministic, given the same seed.
    First the identifiers are sorted before shuffled using the given seed.

    Args:
        identifiers (dict): A dictionary containing the weights for each identifier (key). Per
                            item a dictionary of weights per category is given.
        proportions (dict): Dict of proportions, with a identifier as key.
        seed (int): Seed to use for random operations.

    Returns:
        dict: Dictionary containing a list of identifiers per part with the same key as the proportions dict.

    Example::

        >>> identifiers = {
        >>>     'a': {'music': 2, 'speech': 1},
        >>>     'b': {'music': 5, 'speech': 2},
        >>>     'c': {'music': 2, 'speech': 4},
        >>>     'd': {'music': 1, 'speech': 4},
        >>>     'e': {'music': 3, 'speech': 4}
        >>> }
        >>> proportions = {
        >>>     "train" : 0.6,
        >>>     "dev" : 0.2,
        >>>     "test" : 0.2
        >>> }
        >>> get_identifiers_splitted_by_weights(identifiers, proportions)
        {
            'train': ['a', 'b', 'd'],
            'dev': ['c'],
            'test': ['e']
        }
    """

    identifier_keys = sorted(identifiers.keys())

    # Shuffle to get random order
    rand = random.Random()
    rand.seed(a=seed)
    rand.shuffle(identifier_keys)

    # Get total weight per category
    sum_per_category = collections.defaultdict(int)

    for identifier, cat_weights in identifiers.items():
        for category, weight in cat_weights.items():
            sum_per_category[category] += weight

    target_weights_per_part = collections.defaultdict(dict)

    # Get target weight for each part and category
    for category, total_weight in sum_per_category.items():
        abs_proportions = absolute_proportions(proportions, total_weight)

        for idx, proportion in abs_proportions.items():
            target_weights_per_part[idx][category] = proportion

    print(target_weights_per_part)

    # Distribute items greedily
    part_ids = sorted(list(proportions.keys()))
    current_weights_per_part = {
        idx: collections.defaultdict(int)
        for idx in part_ids
    }
    result = collections.defaultdict(list)

    for identifier in identifier_keys:
        cat_weights = identifiers[identifier]

        target_part = None
        current_part = 0
        weight_over_target = collections.defaultdict(int)

        # Search for fitting part
        while target_part is None and current_part < len(part_ids):
            free_space = True
            part_id = part_ids[current_part]
            part_weights = current_weights_per_part[part_id]

            for category, weight in cat_weights.items():
                target_weight = target_weights_per_part[part_id][category]
                current_weight = part_weights[category]
                weight_diff = target_weight - current_weight - weight
                weight_over_target[part_id] += weight_diff

                if weight_diff < 0:
                    free_space = False

            # If weight doesn't exceed target, place identifier in part
            if free_space:
                target_part = part_id

            current_part += 1

        # If not found fitting part, select the part with the least overweight
        if target_part is None:
            target_part = sorted(
                weight_over_target.items(),
                key=lambda x: x[1],
                reverse=True
            )[0][0]

        result[target_part].append(identifier)

        for category, weight in cat_weights.items():
            current_weights_per_part[target_part][category] += weight

    return result


def select_balanced_subset(items, select_count, categories, select_count_values=None, seed=None):
    """
    Select items so the summed category weights are balanced.
    Each item has a dictionary containing the category weights.
    Items are selected until ``select_count`` is reached.
    The value that is added to ``select_count`` for an item can be defined in the dictionary ``select_count_values``.
    If this is not defined it is assumed to be 1, which means `select_count` items are selected.

    Args:
        items (dict): Dictionary containing items with category weights.
        select_count (float): Value to reach for selected items.
        categories (list): List of all categories.
        select_count_values (dict): The select_count values to be used.
                                    For example an utterance with multiple labels:
                                    The category weights (label-lengths) are used for balance,
                                    but the utterance-duration is used for reaching the select_count.

    Returns:
        list: List of item ids, containing ``number_of_items`` (or ``len(items)`` if smaller).

    Example:
        >>> items = {
        >>>    'utt-1' : {'m': 1, 's': 0, 'n': 0},
        >>>    'utt-2' : {'m': 0, 's': 2, 'n': 1},
        >>>    ...
        >>> }
        >>> select_balanced_subset(items, 5)
        >>> ['utt-1', 'utt-3', 'utt-9', 'utt-33', 'utt-34']
    """

    rand = random.Random()
    rand.seed(seed)

    if select_count_values is None:
        select_count_values = {item_id: 1 for item_id in items.keys()}

    if sum(select_count_values.values()) < select_count:
        return list(items.keys())

    available_item_ids = sorted(list(items.keys()))
    weight_per_category = np.zeros(len(categories))
    selected_item_ids = []
    available_item_weights = []
    current_select_count = 0

    rand.shuffle(available_item_ids)

    # Create dict with weights as vectors
    for item_id in available_item_ids:
        weights = items[item_id]
        all_weights = np.zeros(len(categories))

        for category, weight in weights.items():
            all_weights[categories.index(category)] = float(weight)

        available_item_weights.append(all_weights)

    # Always add best next item
    while current_select_count < select_count:
        best_item_index = 0
        best_item_id = None
        best_item_dist = float('inf')
        current_item_index = 0

        while current_item_index < len(available_item_ids) and best_item_dist > 0:
            item_id = available_item_ids[current_item_index]
            item_weights = available_item_weights[current_item_index]
            temp_total_weights = weight_per_category + item_weights

            dist = temp_total_weights.var()

            if dist < best_item_dist:
                best_item_index = current_item_index
                best_item_dist = dist
                best_item_id = item_id

            current_item_index += 1

        weight_per_category += available_item_weights[best_item_index]
        selected_item_ids.append(best_item_id)
        del available_item_ids[best_item_index]
        del available_item_weights[best_item_index]
        current_select_count += select_count_values[best_item_id]

    return selected_item_ids



import abc
import collections
import copy
import math

import numpy as np

from audiomate.utils import stats


class CorpusView(metaclass=abc.ABCMeta):
    """
    This class defines the basic interface of a corpus. It is not meant to be instantiated directly.
    It only describes the methods for accessing data of the corpus.

    Notes:
        All paths to files should be held as absolute paths in memory.
    """

    @property
    @abc.abstractmethod
    def name(self):
        """ Return the name of the dataset (Equals basename of the path, if not None). """
        return 'undefined'

    #
    #   Tracks
    #

    @property
    @abc.abstractmethod
    def tracks(self):
        """
        Return the tracks in the corpus.

        Returns:
            dict: A dictionary containing :py:class:`audiomate.track.Track` objects with the
            track-idx as key.
        """
        return {}

    @property
    def num_tracks(self):
        """ Return number of tracks. """
        return len(self.tracks)

    def contains_track(self, track):
        """
        Return ``True`` if the given track is in the corpus already,
        ``False`` otherwise.
        """
        if track.idx not in self.tracks.keys():
            return False

        if track != self.tracks[track.idx]:
            return False

        return True

    #
    #   Utterances
    #

    @property
    @abc.abstractmethod
    def utterances(self):
        """
        Return the utterances in the corpus.

        Returns:
            dict: A dictionary containing :py:class:`audiomate.corpus.assets.Utterance` objects with the
            utterance-idx as key.
        """
        return {}

    @property
    def num_utterances(self):
        """ Return number of utterances. """
        return len(self.utterances)

    #
    #   Issuers
    #

    @property
    @abc.abstractmethod
    def issuers(self):
        """
        Return the issuers in the corpus.

        Returns:
            dict: A dictionary containing :py:class:`audiomate.issuers.Issuer` objects with the
            issuer-idx as key.
        """
        return {}

    @property
    def num_issuers(self):
        """ Return the number of issuers in the corpus. """
        return len(self.issuers)

    def contains_issuer(self, issuer):
        """
        Return ``True`` if the given issuer is in the corpus already,
        ``False`` otherwise.
        """
        if issuer.idx not in self.issuers.keys():
            return False

        if issuer != self.issuers[issuer.idx]:
            return False

        return True

        #
        #   Feature Container
        #

    @property
    @abc.abstractmethod
    def feature_containers(self):
        """
        Return the feature-containers in the corpus.

        Returns:
            dict: A dictionary containing :py:class:`audiomate.container.FeatureContainer` objects
            with the feature-idx as key.
        """
        return {}

    @property
    def num_feature_containers(self):
        """ Return the number of feature-containers in the corpus. """
        return len(self.feature_containers)

    #
    #   Subviews
    #

    @property
    def subviews(self):
        """
        Return the subviews of the corpus.

        Returns:
             dict: A dictionary containing :py:class:`audiomate.corpus.Subview` objects with the subview-idx as key.
        """
        return {}

    @property
    def num_subviews(self):
        """ Return the number of subviews in the corpus. """
        return len(self.subviews)

    #
    #   Labels
    #

    def all_label_values(self, label_list_ids=None):
        """
        Return a set of all label-values occurring in this corpus.

        Args:
            label_list_ids (list): If not None, only labels from label-lists with an id contained in this list
                                   are considered.

        Returns:
             :class:`set`: A set of distinct label-values.
        """
        values = set()

        for utterance in self.utterances.values():
            values = values.union(utterance.all_label_values(label_list_ids=label_list_ids))

        return values

    def label_count(self, label_list_ids=None):
        """
        Return a dictionary containing the number of times, every label-value in this corpus is occurring.

        Args:
            label_list_ids (list): If not None, only labels from label-lists with an id contained in this list
                                   are considered.

        Returns:
            dict: A dictionary containing the number of occurrences with the label-value as key.
        """
        count = collections.defaultdict(int)

        for utterance in self.utterances.values():
            for label_value, utt_count in utterance.label_count(label_list_ids=label_list_ids).items():
                count[label_value] += utt_count

        return count

    def label_durations(self, label_list_ids=None):
        """
        Return a dictionary containing the total duration, every label-value in this corpus is occurring.

        Args:
            label_list_ids (list): If not None, only labels from label-lists with an id contained in this list
                                   are considered.

        Returns:
            dict: A dictionary containing the total duration with the label-value as key.
        """
        duration = collections.defaultdict(int)

        for utterance in self.utterances.values():
            for label_value, utt_count in utterance.label_total_duration(label_list_ids=label_list_ids).items():
                duration[label_value] += utt_count

        return duration

    def all_tokens(self, delimiter=' ', label_list_ids=None):
        """
        Return a list of all tokens occurring in one of the labels in the corpus.

        Args:
            delimiter (str): The delimiter used to split labels into tokens
                             (see :meth:`audiomate.annotations.Label.tokenized`).
            label_list_ids (list): If not None, only labels from label-lists with an idx contained in this list
                                   are considered.

        Returns:
             :class:`set`: A set of distinct tokens.
        """
        tokens = set()

        for utterance in self.utterances.values():
            tokens = tokens.union(utterance.all_tokens(delimiter=delimiter, label_list_ids=label_list_ids))

        return tokens

    #
    #   Data
    #

    @property
    def total_duration(self):
        """
        Return the total amount of audio summed over all utterances in the corpus in seconds.
        """
        duration = 0

        for utterance in self.utterances.values():
            duration += utterance.duration

        return duration

    def stats(self):
        """
        Return statistics calculated overall samples of all utterances in the corpus.

        Returns:
            DataStats: A DataStats object containing statistics overall samples in the corpus.
        """

        per_utt_stats = self.stats_per_utterance()
        return stats.DataStats.concatenate(per_utt_stats.values())

    def stats_per_utterance(self):
        """
        Return statistics calculated for all samples of each utterance in the corpus.

        Returns:
            dict: A dictionary containing a DataStats object for each utt.
        """

        all_stats = {}

        for utterance in self.utterances.values():
            data = utterance.read_samples()
            all_stats[utterance.idx] = stats.DataStats(float(np.mean(data)),
                                                       float(np.var(data)),
                                                       np.min(data),
                                                       np.max(data),
                                                       data.size)

        return all_stats

    #
    # Restructuring
    #

    def split_utterances_to_max_time(self, max_time=60.0, overlap=0.0):
        """
        Create a new corpus, where all the utterances are of given maximal duration.
        Utterance longer than ``max_time`` are split up into multiple utterances.

        .. warning::
            Subviews and FeatureContainers are not added to the newly create corpus.

        Arguments:
            max_time (float): Maximal duration for target utterances in seconds.
            overlap (float): Amount of overlap in seconds.
                             The overlap is measured from the center of the splitting.
                             (The actual overlap of two segments is 2 * overlap)

        Returns:
            Corpus: A new corpus instance.
        """

        from audiomate.corpus import Corpus

        result = Corpus()

        # Copy Tracks
        tracks = copy.deepcopy(list(self.tracks.values()))
        result.import_tracks(tracks)

        # Copy Issuers
        issuers = copy.deepcopy(list(self.issuers.values()))
        result.import_issuers(issuers)

        for utterance in self.utterances.values():
            orig_dur = utterance.duration

            if orig_dur > max_time:
                # Compute times where the utterance is split
                num_sub_utts = math.ceil(orig_dur / max_time)
                sub_utt_dur = orig_dur / num_sub_utts
                cutting_points = []

                for i in range(1, num_sub_utts):
                    cutting_points.append(i * sub_utt_dur)

                sub_utts = utterance.split(cutting_points, overlap=overlap)

                # Set track/issuer from new corpus
                for sub_utt in sub_utts:
                    sub_utt.track = result.tracks[utterance.track.idx]

                    if utterance.issuer is not None:
                        sub_utt.issuer = result.issuers[utterance.issuer.idx]

                result.import_utterances(sub_utts)

            # If utterance <= max_time, just copy
            else:
                new_utt = copy.deepcopy(utterance)
                new_utt.track = result.tracks[new_utt.track.idx]

                if new_utt.issuer is not None:
                    new_utt.issuer = result.issuers[new_utt.issuer.idx]

                result.import_utterances(new_utt)

        return result



import os
import numpy as np

from audiomate import containers

import pytest
from tests import resources


@pytest.fixture()
def sample_container():
    container_path = resources.get_resource_path(
        ['sample_files', 'audio_container']
    )
    sample_container = containers.AudioContainer(container_path)
    sample_container.open()
    yield sample_container
    sample_container.close()


class TestAudioContainer:

    def test_get(self, sample_container):
        samples, sr = sample_container.get('track1')

        assert np.allclose(
            samples,
            np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]),
            atol=1.e-4
        )
        assert sr == 16000

    def test_set(self, tmpdir):
        path = os.path.join(tmpdir.strpath, 'audio')
        cnt = containers.AudioContainer(path)
        cnt.open()

        data = np.random.random(10)
        cnt.set('track1', data, 16000)

        assert cnt.keys() == ['track1']

        samples, sr = cnt.get('track1')

        assert np.allclose(samples, data, atol=1.e-4)
        assert sr == 16000

        cnt.close()

    def test_set_raises_error_on_invalid_data_type(self, tmpdir):
        path = os.path.join(tmpdir.strpath, 'audio')
        cnt = containers.AudioContainer(path)
        cnt.open()

        with pytest.raises(ValueError):
            cnt.set('track1', np.arange(5).astype(np.int16), 16000)

        cnt.close()

    def test_set_raises_error_on_multi_dim_data(self, tmpdir):
        path = os.path.join(tmpdir.strpath, 'audio')
        cnt = containers.AudioContainer(path)
        cnt.open()

        with pytest.raises(ValueError):
            cnt.set('track1',
                    np.random.random((4, 5)).astype(np.float32),
                    16000)

        cnt.close()

    def test_append(self, tmpdir):
        path = os.path.join(tmpdir.strpath, 'audio')
        cnt = containers.AudioContainer(path)
        cnt.open()

        chunk_a = np.random.random(10).astype(np.float32)
        cnt.append('track1', np.array(chunk_a), 16000)
        samples, sr = cnt.get('track1')

        assert np.allclose(samples, chunk_a, atol=1.e-4)
        assert sr == 16000

        chunk_b = np.random.random(5).astype(np.float32)
        cnt.append('track1', np.array(chunk_b), 16000)
        samples, sr = cnt.get('track1')

        assert np.allclose(
            samples,
            np.concatenate([chunk_a, chunk_b]),
            atol=1.e-4
        )
        assert sr == 16000

        cnt.close()



import os

import audiomate
from audiomate import annotations
from audiomate.utils import download
from audiomate.utils import files
from . import base

DOWNLOAD_URL = 'http://opihi.cs.uvic.ca/sound/music_speech.tar.gz'
DIRECTORIES = {'music_wav': 'music', 'speech_wav': 'speech'}


class GtzanDownloader(base.CorpusDownloader):
    """
    Downloader for the GTZAN Corpus.

    Args:
        url (str): The url to download the dataset from. If not given the default URL is used.
                   It is expected to be a tar.gz file.
    """

    def __init__(self, url=None):
        if url is None:
            self.url = DOWNLOAD_URL
        else:
            self.url = url

    @classmethod
    def type(cls):
        return 'gtzan'

    def _download(self, target_path):
        os.makedirs(target_path, exist_ok=True)
        tmp_file = os.path.join(target_path, 'tmp_ark.tar.gz')

        download.download_file(self.url, tmp_file)
        download.extract_tar(tmp_file, target_path)

        # We use copy since subfolders in the archive are read-only, hence throws permission error when trying to move.
        files.move_all_files_from_subfolders_to_top(target_path, delete_subfolders=True, copy=True)

        os.remove(tmp_file)


class GtzanReader(base.CorpusReader):
    """
    Reader for the GTZAN music/speech corpus. The corpus consits of 64 music and 64 speech tracks that are each 30
    seconds long. The Wave files are 16-bit mono and have a sampling rate of 22050 Hz.

    .. seealso::

       `MARSYAS: GTZAN Music/Speech <https://marsyasweb.appspot.com/download/data_sets/>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'gtzan'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        for directory, type_name in DIRECTORIES.items():
            source_directory = os.path.join(path, directory)

            if not os.path.isdir(source_directory):
                continue

            it = os.scandir(source_directory)

            for entry in it:
                if not entry.name.endswith('.wav'):
                    continue

                file_path = os.path.join(source_directory, entry.name)
                file_idx = entry.name[0:-4]  # chop of .wav
                utterance_idx = file_idx  # every file is a separate utterance

                corpus.new_file(file_path, track_idx=file_idx, copy_file=False)
                utterance = corpus.new_utterance(utterance_idx, file_idx)
                utterance.set_label_list(annotations.LabelList.create_single(type_name, idx=audiomate.corpus.LL_DOMAIN))

        return corpus



from audiomate.corpus import validation


class TestInvalidItemsResult:

    def test_get_report_on_pass(self):
        result = validation.InvalidItemsResult(passed=True, invalid_items={}, name='MyName', info={'a': 'b'})

        assert result.get_report() == '\n'.join([
            'MyName',
            '======',
            '',
            '--> a: b',
            '',
            'Result: Passed'
        ])

    def test_get_report_on_fail(self):
        result = validation.InvalidItemsResult(passed=False, invalid_items={
            'utt-1': 'reason1',
            'utt-4': 'reason4',
            'utt-2': 3
        }, name='MyName', info={'c': 'b', 'a': 49.0})

        assert result.get_report() == '\n'.join([
            'MyName',
            '======',
            '',
            '--> a: 49.0',
            '--> c: b',
            '',
            'Result: Failed',
            '',
            'Invalid Utterances:',
            '    * utt-1 (reason1)',
            '    * utt-2 (3)',
            '    * utt-4 (reason4)'
        ])

    def test_get_report_on_fail_with_type_name(self):
        result = validation.InvalidItemsResult(
            passed=False,
            invalid_items={
                'utt-1': 'reason1',
                'utt-4': 'reason4',
                'utt-2': 3
            },
            name='MyName',
            item_name='Tracks',
            info={'c': 'b', 'a': 49.0}
        )

        assert result.get_report() == '\n'.join([
            'MyName',
            '======',
            '',
            '--> a: 49.0',
            '--> c: b',
            '',
            'Result: Failed',
            '',
            'Invalid Tracks:',
            '    * utt-1 (reason1)',
            '    * utt-2 (3)',
            '    * utt-4 (reason4)'
        ])



from . import base


class CombinedValidationResult(base.ValidationResult):
    """
    Result of running multiple validation-tasks with the validator.

    Args:
        passed (bool): A boolean, indicating if all tasks have passed (True) or at least one failed (False).
        results (dict): A dictionary containing the results of all validators, with the task name as key.
        info (dict): Dictionary containing key/value string-pairs with detailed information of the validation.
                     For example id of the label-list that was validated.
    """

    def __init__(self, passed, results=None, info=None):
        super(CombinedValidationResult, self).__init__(passed, info=info)

        self.results = results or {}

    def get_report(self):
        sorted_val = sorted(self.results.items(), key=lambda x: x[0])
        lines = ['{} --> {}'.format(x, 'Passed' if y.passed else 'Failed') for x, y in sorted_val]
        lines.append('\n')

        for name in sorted(self.results.keys()):
            result = self.results[name]

            lines.append(result.get_report())
            lines.append('\n')

        return '\n'.join(lines)


class CombinedValidator(base.Validator):
    """
    The CombinedValidator is used to execute multiple validators at once.

    Args:
        validators (list): A list of validators that are executed.
    """

    def __init__(self, validators=None):
        self.validators = validators or []

    def name(cls):
        return 'Combined-Validator'

    def validate(self, corpus):
        """
        Perform validation on the given corpus.

        Args:
            corpus (Corpus): The corpus to test/validate.
        """

        passed = True
        results = {}

        for validator in self.validators:
            sub_result = validator.validate(corpus)
            results[validator.name()] = sub_result

            if not sub_result.passed:
                passed = False

        return CombinedValidationResult(passed, results)



import os

import pytest
import requests_mock

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus import io
from audiomate.corpus.io import mailabs

from tests import resources
from . import reader_test as rt


@pytest.fixture()
def tar_data():
    with open(resources.get_resource_path(['sample_files', 'cv_corpus_v1.tar.gz']), 'rb') as f:
        return f.read()


class TestMailabsDownloader:

    def test_download(self, tar_data, tmpdir):
        target_folder = tmpdir.strpath
        downloader = mailabs.MailabsDownloader(tags='de_DE')

        with requests_mock.Mocker() as mock:
            mock.get(mailabs.DOWNLOAD_URLS['de_DE'], content=tar_data)
            downloader.download(target_folder)

        base_path = os.path.join(target_folder, 'common_voice')

        assert os.path.isfile(os.path.join(base_path, 'cv-valid-dev.csv'))
        assert os.path.isdir(os.path.join(base_path, 'cv-valid-dev'))
        assert os.path.isfile(os.path.join(base_path, 'cv-valid-train.csv'))
        assert os.path.isdir(os.path.join(base_path, 'cv-valid-train'))


class TestMailabsReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('mailabs')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'files')

    EXPECTED_NUMBER_OF_TRACKS = 13
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('elizabeth_klett-jane_eyre_01_f000001',
                        path=os.path.join(SAMPLE_PATH, 'en_US', 'by_book', 'female',
                                          'elizabeth_klett', 'jane_eyre',
                                          'wavs', 'jane_eyre_01_f000001.wav')),
        rt.ExpFileTrack('fred-azele_01_f000002',
                        path=os.path.join(SAMPLE_PATH, 'de_DE', 'by_book', 'male', 'fred',
                                          'azele', 'wavs', 'azele_01_f000002.wav')
                        ),
        rt.ExpFileTrack('abc_01_f000002',
                        path=os.path.join(SAMPLE_PATH, 'de_DE', 'by_book', 'mix',
                                          'abc', 'wavs', 'abc_01_f000002.wav')
                        ),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 6
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('fred', 5, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('elizabeth_klett', 3, issuers.Gender.FEMALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('abc_01_f000002', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 13
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('elizabeth_klett-jane_eyre_01_f000003', 'elizabeth_klett-jane_eyre_01_f000003',
                        'elizabeth_klett', 0, float('inf')),
        rt.ExpUtterance('sara-abc_01_f000001', 'sara-abc_01_f000001', 'sara', 0, float('inf')),
        rt.ExpUtterance('abc_01_f000001', 'abc_01_f000001', 'abc_01_f000001', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        'elizabeth_klett-jane_eyre_01_f000003': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1),
        ],
        'abc_01_f000002': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1),
        ],
    }

    EXPECTED_LABELS = {
        'elizabeth_klett-jane_eyre_01_f000003': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT_RAW, 'Chapter 1.', 0, float('inf')),
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Chapter one.', 0, float('inf')),
        ],
        'abc_01_f000002': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT_RAW, 'das 1.', 0, float('inf')),
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'das eins.', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 2
    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('de_DE', [
            'fred-thego_01_f000001',
            'fred-thego_01_f000002',
            'fred-thego_01_f000003',
            'fred-azele_01_f000001',
            'fred-azele_01_f000002',
            'tim-azele_01_f000001',
            'sara-abc_01_f000001',
            'sara-abc_01_f000002',
            'abc_01_f000001',
            'abc_01_f000002',
        ]),
        rt.ExpSubview('en_US', [
            'elizabeth_klett-jane_eyre_01_f000001',
            'elizabeth_klett-jane_eyre_01_f000002',
            'elizabeth_klett-jane_eyre_01_f000003',
        ]),
    ]

    def load(self):
        return io.MailabsReader().load(self.SAMPLE_PATH)



import os

import numpy as np
import librosa

from audiomate import tracks
from audiomate.utils import audio


def test_read_blocks(tmpdir):
    wav_path = os.path.join(tmpdir.strpath, 'file.wav')
    wav_content = np.random.random(10000)
    librosa.output.write_wav(wav_path, wav_content, 16000)

    blocks = list(audio.read_blocks(wav_path, buffer_size=1000))

    assert np.allclose(np.concatenate(blocks), wav_content, atol=0.0001)
    assert np.concatenate(blocks).dtype == np.float32


def test_read_blocks_with_start_end(tmpdir):
    wav_path = os.path.join(tmpdir.strpath, 'file.wav')
    wav_content = np.random.random(10000)
    librosa.output.write_wav(wav_path, wav_content, 16000)

    blocks = list(audio.read_blocks(wav_path, start=0.1, end=0.3, buffer_size=1000))

    assert np.concatenate(blocks).dtype == np.float32
    assert np.allclose(np.concatenate(blocks), wav_content[1600:4800], atol=0.0001)


def test_read_frames(tmpdir):
    wav_path = os.path.join(tmpdir.strpath, 'file.wav')
    wav_content = np.random.random(10044)
    librosa.output.write_wav(wav_path, wav_content, 16000)

    data = list(audio.read_frames(wav_path, frame_size=400, hop_size=160))
    frames = np.array([x[0] for x in data])
    last = [x[1] for x in data]

    assert frames.shape == (62, 400)
    assert frames.dtype == np.float32
    assert np.allclose(frames[0], wav_content[:400], atol=0.0001)
    assert np.allclose(frames[61], np.pad(wav_content[9760:], (0, 116), mode='constant'), atol=0.0001)

    assert last[:-1] == [False] * (len(data) - 1)
    assert last[-1]


def test_read_frames_matches_length(tmpdir):
    wav_path = os.path.join(tmpdir.strpath, 'file.wav')
    wav_content = np.random.random(10000)
    librosa.output.write_wav(wav_path, wav_content, 16000)

    data = list(audio.read_frames(wav_path, frame_size=400, hop_size=160))
    frames = np.array([x[0] for x in data])
    last = [x[1] for x in data]

    assert frames.shape == (61, 400)
    assert frames.dtype == np.float32
    assert np.allclose(frames[0], wav_content[:400], atol=0.0001)
    assert np.allclose(frames[60], wav_content[9600:], atol=0.0001)

    assert last[:-1] == [False] * (len(data) - 1)
    assert last[-1]


def test_write_wav(tmpdir):
    samples = np.random.random(50000)
    sr = 16000
    path = os.path.join(tmpdir.strpath, 'audio.wav')

    audio.write_wav(path, samples, sr=sr)

    assert os.path.isfile(path)

    track = tracks.FileTrack('idx', path)
    assert np.allclose(
        samples,
        track.read_samples(),
        atol=1.e-04
    )



import os

from audiomate import corpus
from audiomate.corpus import io

from tests import resources
from . import reader_test as rt


class TestUrbansound8kReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('urbansound8k')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'audio')

    EXPECTED_NUMBER_OF_TRACKS = 5
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('100032-3-0-0', os.path.join('fold5', '100032-3-0-0.wav')),
        rt.ExpFileTrack('100263-2-0-117', os.path.join('fold5', '100263-2-0-117.wav')),
        rt.ExpFileTrack('145612-6-3-0', os.path.join('fold8', '145612-6-3-0.wav')),
        rt.ExpFileTrack('145683-6-5-0', os.path.join('fold9', '145683-6-5-0.wav')),
        rt.ExpFileTrack('79377-9-0-4', os.path.join('fold2', '79377-9-0-4.wav')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 0

    EXPECTED_NUMBER_OF_UTTERANCES = 5
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('100032-3-0-0', '100032-3-0-0', None, 0, float('inf')),
        rt.ExpUtterance('100263-2-0-117', '100263-2-0-117', None, 0, float('inf')),
        rt.ExpUtterance('145612-6-3-0', '145612-6-3-0', None, 0, float('inf')),
        rt.ExpUtterance('145683-6-5-0', '145683-6-5-0', None, 0, float('inf')),
        rt.ExpUtterance('79377-9-0-4', '79377-9-0-4', None, 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        '100032-3-0-0': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
        '100263-2-0-117': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
        '145612-6-3-0': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
        '145683-6-5-0': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
        '79377-9-0-4': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
    }

    EXPECTED_LABELS = {
        '100032-3-0-0':   [rt.ExpLabel(corpus.LL_SOUND_CLASS, 'dog_bark', 0, float('inf'))],
        '100263-2-0-117': [rt.ExpLabel(corpus.LL_SOUND_CLASS, 'children_playing', 0, float('inf'))],
        '145612-6-3-0':   [rt.ExpLabel(corpus.LL_SOUND_CLASS, 'gun_shot', 0, float('inf'))],
        '145683-6-5-0':   [rt.ExpLabel(corpus.LL_SOUND_CLASS, 'gun_shot', 0, float('inf'))],
        '79377-9-0-4':    [rt.ExpLabel(corpus.LL_SOUND_CLASS, 'street_music', 0, float('inf'))],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 4
    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('fold2', [
            '79377-9-0-4',
        ]),
        rt.ExpSubview('fold5', [
            '100032-3-0-0',
            '100263-2-0-117',
        ]),
        rt.ExpSubview('fold8', [
            '145612-6-3-0',
        ]),
        rt.ExpSubview('fold9', [
            '145683-6-5-0',
        ]),
    ]

    def load(self):
        return io.Urbansound8kReader().load(self.SAMPLE_PATH)



import click
import h5py
import numpy as np

from audiomate.feeding import PartitioningFeatureIterator


@click.command()
@click.option('--h5-file', help='h5 file with features to read', type=click.Path(exists=True), required=True)
@click.option('--partition/--no-partition', default=False)
@click.option('--shuffle/--no-shuffle', default=False)
@click.option('--partition-size', default='12g')
def run(h5_file, partition, shuffle, partition_size):
    file = h5py.File(h5_file, 'r')
    counter = 0
    features_bytes_read = 0

    num_of_features = 0
    dsets = {}
    for dset in file:
        length = len(file[dset])
        dsets[dset] = length
        num_of_features += length

    if partition:
        for feature in PartitioningFeatureIterator(file, partition_size, shuffle=shuffle):
            features_bytes_read += feature[2].dtype.itemsize * len(feature[2])
            counter += 1
    else:
        if shuffle:
            indices = np.random.permutation(num_of_features)

            for idx in indices:
                for dset, length in dsets.items():
                    if idx >= length:
                        idx -= length
                        continue

                    feature = file[dset][idx]
                    features_bytes_read += feature.dtype.itemsize * len(feature)
                    counter += 1
                    break
        else:
            for dset in file:
                for feature in file[dset]:
                    features_bytes_read += feature.dtype.itemsize * len(feature)
                    counter += 1

    print('Total number of features: {0}'.format(num_of_features))
    print('Features read: {0}'.format(counter))
    print('Bytes of features read: {0}'.format(features_bytes_read))


if __name__ == '__main__':
    run()



import os
import glob

import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.corpus import subset
from . import base
from audiomate.utils import textfile


class TimitReader(base.CorpusReader):
    """
    Reader for the TIMIT Corpus.

    .. seealso::

       `TIMIT <https://github.com/philipperemy/timit>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'timit'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        for part in ['TEST', 'TRAIN']:
            part_path = os.path.join(path, part)
            part_utt_ids = set()

            for region in os.listdir(part_path):
                region_path = os.path.join(part_path, region)

                if os.path.isdir(region_path):

                    for speaker_abbr in os.listdir(region_path):
                        speaker_path = os.path.join(region_path, speaker_abbr)
                        speaker_idx = speaker_abbr[1:]

                        if speaker_idx not in corpus.issuers.keys():
                            issuer = issuers.Speaker(speaker_idx)

                            if speaker_abbr[:1] == 'M':
                                issuer.gender = issuers.Gender.MALE
                            elif speaker_abbr[:1] == 'F':
                                issuer.gender = issuers.Gender.FEMALE

                            corpus.import_issuers(issuer)

                        for wav_path in glob.glob(os.path.join(speaker_path, '*.WAV')):
                            sentence_idx = os.path.splitext(os.path.basename(wav_path))[0]
                            utt_idx = '{}-{}-{}'.format(region, speaker_abbr, sentence_idx).lower()
                            part_utt_ids.add(utt_idx)

                            raw_text_path = os.path.join(speaker_path, '{}.TXT'.format(sentence_idx))
                            raw_text = textfile.read_separated_lines(raw_text_path, separator=' ', max_columns=3)[0][2]

                            words_path = os.path.join(speaker_path, '{}.WRD'.format(sentence_idx))
                            words = textfile.read_separated_lines(words_path, separator=' ', max_columns=3)

                            phones_path = os.path.join(speaker_path, '{}.PHN'.format(sentence_idx))
                            phones = textfile.read_separated_lines(phones_path, separator=' ', max_columns=3)

                            corpus.new_file(wav_path, utt_idx)
                            utt = corpus.new_utterance(utt_idx, utt_idx, speaker_idx)

                            raw_ll = annotations.LabelList.create_single(raw_text,
                                                                         idx=audiomate.corpus.LL_WORD_TRANSCRIPT_RAW)
                            utt.set_label_list(raw_ll)

                            word_ll = annotations.LabelList(idx=audiomate.corpus.LL_WORD_TRANSCRIPT)

                            for record in words:
                                start = int(record[0]) / 16000
                                end = int(record[1]) / 16000
                                word_ll.addl(record[2], start=start, end=end)

                            utt.set_label_list(word_ll)

                            phone_ll = annotations.LabelList(idx=audiomate.corpus.LL_PHONE_TRANSCRIPT)

                            for record in phones:
                                start = int(record[0]) / 16000
                                end = int(record[1]) / 16000
                                phone_ll.addl(record[2], start=start, end=end)

                            utt.set_label_list(phone_ll)

            filter = subset.MatchingUtteranceIdxFilter(utterance_idxs=part_utt_ids)
            subview = subset.Subview(corpus, filter_criteria=[filter])
            corpus.import_subview(part, subview)

        return corpus



"""
This module module provides functionality to load data from a container into memory in chunks.
"""

import gc
import random

import numpy as np

import audiomate
from audiomate import containers
from audiomate.utils import units


class PartitioningContainerLoader(object):
    """
    Load data from one or more containers in partitions.
    It computes a scheme to load the data of as many utterances as possible in one partition.

    A scheme is initially computed on creation of the loader. To compute a new one the ``reload()`` method can be used.
    This only has an effect if ``shuffle == True``,
    otherwise the utterances are defined always loaded in the same order.

    With a given scheme, data of a partition can be retrieved via ``load_partition_data()``.
    It loads all data of the partition with the given index into memory.

    Args:
        corpus_or_utt_ids (Corpus, list): Either a corpus or a list of utterances.
                                          This defines which utterances are considered for loading.
        containers (container.Container, list): Either a single or a list of Container objects.
                                             From the given containers data is loaded.
        partition_size (str): Size of the partitions in bytes. The units ``k`` (kibibytes), ``m`` (mebibytes) and ``g``
                             (gibibytes) are supported, i.e. a ``partition_size`` of ``1g`` equates :math:`2^{30}`
                             bytes.
        shuffle (bool): Indicates whether the utterances should be returned in
                        random order (``True``) or not (``False``).
        seed (int): Seed to be used for the random number generator.

    Example:
        >>> corpus = audiomate.Corpus.load('/path/to/corpus')
        >>> container_inputs = containers.FeatureContainer('/path/to/features.hdf5')
        >>> container_outputs = containers.Container('/path/to/targets.hdf5')
        >>>
        >>> lo = PartitioningContainerLoader(corpus, [container_inputs, container_outputs], '1G', shuffle=True, seed=23)
        >>> len(lo.partitions) # Number of parititions
        5
        >>> lo.partitions[0].utt_ids # Utterances in the partition with index 0
        ['utt-1', 'utt-2', ...]
        >>> p0 = lo.load_partition_data(0) # Load partition 0 into memory
        >>> p0.info.utt_ids[0] # First utterance in the partition
        'utt-1'
        >>> p0.utt_data[0] # Data of the first utterance
        (
            array([[0.58843831, 0.18128443, 0.19718328, 0.25284105], ...]),
            array([[0.0, 1.0], ...])
        )
    """

    def __init__(self, corpus_or_utt_ids, feature_containers, partition_size,
                 shuffle=True, seed=None):
        if isinstance(corpus_or_utt_ids, audiomate.Corpus):
            self.utt_ids = list(corpus_or_utt_ids.utterances.keys())
        else:
            self.utt_ids = corpus_or_utt_ids

        if isinstance(feature_containers, containers.Container):
            self.containers = [feature_containers]
        else:
            self.containers = feature_containers

        if len(self.containers) == 0:
            raise ValueError('At least one container has to be provided!')

        self.partitions = []
        self.partition_size = units.parse_storage_size(partition_size)
        self.shuffle = shuffle

        # init random state
        self.rand = random.Random()
        self.rand.seed(a=seed)

        # check
        self._raise_error_if_container_is_missing_an_utterance()

        # Compute utterance size and length
        self.utt_sizes = self._scan()
        self.utt_lengths = self._get_all_lengths()

        self.reload()

    def reload(self):
        """
        Create a new partition scheme. A scheme defines which utterances are in which partition.
        The scheme only changes after every call if ``self.shuffle == True``.

        Returns:
            list: List of PartitionInfo objects, defining the new partitions (same as ``self.partitions``)
        """

        # Create the order in which utterances will be loaded
        utt_ids = sorted(self.utt_ids)

        if self.shuffle:
            self.rand.shuffle(utt_ids)

        partitions = []

        current_partition = PartitionInfo()

        for utt_id in utt_ids:
            utt_size = self.utt_sizes[utt_id]
            utt_lengths = self.utt_lengths[utt_id]

            # We add utterance to the partition as long the partition-size is not exceeded
            # Otherwise we start with new partition.
            if current_partition.size + utt_size > self.partition_size:
                partitions.append(current_partition)
                current_partition = PartitionInfo()

            current_partition.utt_ids.append(utt_id)
            current_partition.utt_lengths.append(utt_lengths)
            current_partition.size += utt_size

        if current_partition.size > 0:
            partitions.append(current_partition)

        self.partitions = partitions
        return self.partitions

    def load_partition_data(self, index):
        """
        Load and return the partition with the given index.

        Args:
            index (int): The index of partition, that refers to the index in ``self.partitions``.

        Returns:
            PartitionData: A PartitionData object containing the data for the partition with the given index.
        """

        info = self.partitions[index]
        data = PartitionData(info)

        for utt_id in info.utt_ids:
            utt_data = [c._file[utt_id][:] for c in self.containers]
            data.utt_data.append(utt_data)

        return data

    def _raise_error_if_container_is_missing_an_utterance(self):
        """ Check if there is a dataset for every utterance in every container, otherwise raise an error. """
        expected_keys = frozenset(self.utt_ids)

        for cnt in self.containers:
            keys = set(cnt.keys())

            if not keys.issuperset(expected_keys):
                raise ValueError('Container is missing data for some utterances!')

    def _scan(self):
        """ For every utterance, calculate the size it will need in memory. """
        utt_sizes = {}

        for dset_name in self.utt_ids:
            per_container = []

            for cnt in self.containers:
                dset = cnt._file[dset_name]
                dtype_size = dset.dtype.itemsize

                record_size = dtype_size * dset.size
                per_container.append(record_size)

            utt_size = sum(per_container)

            if utt_size > self.partition_size:
                raise ValueError('Records in "{0}" are larger than the partition size'.format(dset_name))

            utt_sizes[dset_name] = utt_size

        return utt_sizes

    def _get_all_lengths(self):
        """ For every utterance, get the length of the data in every container. Return a list of tuples. """
        utt_lengths = {}

        for utt_idx in self.utt_ids:
            per_container = [c._file[utt_idx].shape[0] for c in self.containers]
            utt_lengths[utt_idx] = tuple(per_container)

        return utt_lengths


class PartitionInfo(object):
    """
    Class for holding the info of a partition.

    Attributes:
        utt_ids (list): A list of utterance-ids in the partition.
        utt_lengths (list): List with lengths of the utterances. (Outermost dimension in the dataset of the container)
                            Since there are maybe multiple containers, every item is a tuple of lengths.
                            They correspond to the length of the utterance in every container,
                            in the order of the containers passed to the ParitioningContainerLoader.
        size (int): The number of bytes the partition will allocate, when loaded.
    """

    def __init__(self):
        self.utt_ids = []
        self.utt_lengths = []
        self.size = 0

    def total_lengths(self):
        """ Return the total length of all utterances for every container. """
        return tuple([sum(x) for x in zip(*self.utt_lengths)])


class PartitionData(object):
    """
    Class for holding the loaded data of a partition.

    Args:
        info (PartitionInfo): The info about the partition.

    Attributes:
        utt_data (list): A list holding the data-objects for every utterance in the order of ``info.utt_ids``.
                         The entries are also lists or tuples containing the array for every container.
    """

    def __init__(self, info):
        self.info = info
        self.utt_data = []


class PartitioningFeatureIterator(object):
    """
    Iterates over all features in the given HDF5 file.

    Before iterating over the features, the iterator slices the file into one or more partitions and loads the data into
    memory. This leads to significant speed-ups even with moderate partition sizes, regardless of the type of disk
    (spinning or flash). Pseudo random access is supported with a negligible impact on performance and randomness: The
    data is randomly sampled (without replacement) within each partition and the partitions are loaded in random order,
    too.

    The features are emitted as triplets in the form of
    ``(utterance name, index of the feature within the utterance, feature)``.

    When calculating the partition sizes only the size of the features itself is factored in, overhead of data storage
    is ignored. This overhead is usually negligible even with partition sizes of multiple gigabytes because the data is
    stored as numpy ndarrays in memory (one per utterance). The overhead of a single ndarray is 96 bytes regardless of
    its size. Nonetheless the partition size should be chosen to be lower than the total available memory.

    Args:
        hdf5file(h5py.File): HDF5 file containing the features
        partition_size(str): Size of the partitions in bytes. The units ``k`` (kibibytes), ``m`` (mebibytes) and ``g``
                             (gibibytes) are supported, i.e. a ``partition_size`` of ``1g`` equates :math:`2^{30}`
                             bytes.
        shuffle(bool): Indicates whether the features should be returned in random order (``True``) or not (``False``).
        seed(int): Seed to be used for the random number generator.
        includes(iterable): Iterable of names of data sets that should be included when iterating over the feature
                            container. Mutually exclusive with ``excludes``. If both are specified, only ``includes``
                            will be considered.
        excludes(iterable): Iterable of names of data sets to skip when iterating over the feature container. Mutually
                            exclusive with ``includes``. If both are specified, only ``includes`` will be considered.

    Example:
        >>> import h5py
        >>> from audiomate.feeding import PartitioningFeatureIterator
        >>> hdf5 = h5py.File('features.h5', 'r')
        >>> iterator = PartitioningFeatureIterator(hdf5, '12g', shuffle=True)
        >>> next(iterator)
        ('music-fma-0100', 227, array([-0.15004082, -0.30246958, -0.38708138, ..., -0.93471956,
               -0.94194776, -0.90878332], dtype=float32))
        >>> next(iterator)
        ('music-fma-0081', 2196, array([-0.00207647, -0.00101351, -0.00058832, ..., -0.00207647,
               -0.00292684, -0.00292684], dtype=float32))
        >>> next(iterator)
        ('music-hd-0050', 1026, array([-0.57352495, -0.63049972, -0.63049972, ...,  0.82490814,
                0.84680521,  0.75517786], dtype=float32))
    """

    def __init__(self, hdf5file, partition_size, shuffle=True, seed=None, includes=None, excludes=None):
        self._file = hdf5file
        self._partition_size = units.parse_storage_size(partition_size)
        self._shuffle = shuffle
        self._seed = seed

        data_sets = self._filter_data_sets(hdf5file.keys(), includes=includes, excludes=excludes)
        if shuffle:
            _random_state(self._seed).shuffle(data_sets)

        self._data_sets = tuple(data_sets)
        self._partitions = []
        self._partition_idx = 0
        self._partition_data = None

        self._partition()

    def __iter__(self):
        return self

    def __next__(self):
        if self._partition_data is None or not self._partition_data.has_next():
            if self._partition_data is not None:
                self._partition_data = None
                gc.collect()  # signal gc that it's time to get rid of the obsolete data

            self._partition_data = self._load_next_partition()

            if self._partition_data is None:
                raise StopIteration

        return next(self._partition_data)

    def _load_next_partition(self):
        if len(self._partitions) == self._partition_idx:
            return None

        start, end = self._partitions[self._partition_idx]
        self._partition_idx += 1

        start_dset_name, start_idx = start
        end_dset_name, end_idx = end

        start_dset_idx = self._data_sets.index(start_dset_name)
        end_dset_idx = self._data_sets.index(end_dset_name)

        if start_dset_name == end_dset_name:
            slices = [DataSetSlice(start_dset_name, start_idx, self._file[start_dset_name][start_idx:end_idx])]
            return Partition(slices, shuffle=self._shuffle, seed=self._seed)

        slices = [DataSetSlice(start_dset_name, start_idx, self._file[start_dset_name][start_idx:])]

        middle_dsets = self._data_sets[start_dset_idx + 1:end_dset_idx]
        for dset in middle_dsets:
            slices.append(DataSetSlice(dset, 0, self._file[dset][:]))

        slices.append(DataSetSlice(end_dset_name, 0, self._file[end_dset_name][:end_idx]))

        return Partition(slices, shuffle=self._shuffle, seed=self._seed)

    def _partition(self):
        dset_props = self._scan()

        start = None
        partition_free_space = self._partition_size

        for idx, props in enumerate(dset_props):
            dset_name = props.name
            num_records = props.num_of_records
            record_size = props.record_size
            remaining_records = props.num_of_records
            is_last = (idx == len(dset_props) - 1)

            next_record_size = None if is_last else dset_props[idx + 1].record_size

            if start is None:
                start = (dset_name, 0)

            while partition_free_space >= record_size and remaining_records >= 1:
                num_fitting_records = int(partition_free_space / record_size)
                num_records_taken = min(remaining_records, num_fitting_records)
                end_index = num_records_taken if dset_name != start[0] else start[1] + num_records_taken
                end = (dset_name, end_index)

                if num_records_taken == num_fitting_records:  # Partition is going to be full afterwards
                    self._partitions.append((start, end))

                    partition_free_space = self._partition_size

                    if end[1] == num_records:  # Data set is exhausted
                        start = None
                        break
                    else:  # Next partition starts within the same data set
                        start = end
                elif num_records_taken == remaining_records and is_last:  # All data sets are partitioned
                    self._partitions.append((start, end))
                    break
                else:
                    partition_free_space -= record_size * num_records_taken

                    if partition_free_space < next_record_size:
                        self._partitions.append((start, end))
                        start = None
                        partition_free_space = self._partition_size
                        break

                remaining_records -= num_records_taken

        if self._shuffle:
            _random_state(self._seed).shuffle(self._partitions)

    def _scan(self):
        dset_props = []

        for dset_name in self._data_sets:
            dtype_size = self._file[dset_name].dtype.itemsize

            if len(self._file[dset_name]) == 0:
                continue

            num_records, items_per_record = self._file[dset_name].shape
            record_size = dtype_size * items_per_record

            if record_size > self._partition_size:
                raise ValueError('Records in "{0}" are larger than the partition size'.format(dset_name))

            dset_props.append(DataSetProperties(dset_name, num_records, record_size))

        return dset_props

    @staticmethod
    def _filter_data_sets(data_sets, includes=None, excludes=None):
        if includes is None:
            includes = frozenset()
        else:
            includes = frozenset(includes)

        if excludes is None:
            excludes = frozenset()
        else:
            excludes = frozenset(excludes)

        if len(includes) > 0:
            return [data_set for data_set in data_sets if data_set in includes]

        return [data_set for data_set in data_sets if data_set not in excludes]


class DataSetProperties:
    def __init__(self, name, num_of_records, record_size):
        self.name = name
        self.num_of_records = num_of_records
        self.record_size = record_size

    def __repr__(self):
        return 'DataSetProperties({0}, {1}, {2})'.format(self.name, self.num_of_records, self.record_size)


class Partition:
    def __init__(self, slices, shuffle=True, seed=None):
        self._slices = slices

        self._total_length = 0
        for item in slices:
            self._total_length += item.length

        self._index = 0

        if shuffle:
            self._elements = _random_state(seed).permutation(self._total_length)
        else:
            self._elements = np.arange(0, self._total_length)

    def __iter__(self):
        return self

    def __next__(self):
        if self._index == self._total_length:
            raise StopIteration()

        index = self._elements[self._index]
        for item in self._slices:
            if index >= item.length:
                index -= item.length
                continue

            self._index += 1

            # emits triplet (data set's name, original index of feature within data set, feature)
            return item.data_set_name, item.start_index + index, item.data[index]

    def has_next(self):
        return self._index < self._total_length


class DataSetSlice:
    def __init__(self, data_set_name, start_index, data):
        self.data_set_name = data_set_name
        self.start_index = start_index
        self.length = len(data)
        self.data = data


def _random_state(seed=None):
    random_state = np.random.RandomState()

    if seed is not None:
        random_state.seed(seed)

    return random_state



import os
import tempfile

from audiomate.utils import textfile


def test_read_separated_lines():
    file_path = os.path.join(os.path.dirname(__file__), 'multi_column_file.txt')

    expected = [
        ['a', '1', 'x'],
        ['b', '2', 'y'],
        ['c', '3', 'z']
    ]

    records = textfile.read_separated_lines(file_path, separator='\t')

    assert expected == records


def test_read_separated_lines_with_first_key():
    file_path = os.path.join(os.path.dirname(__file__), 'multi_column_file.txt')

    expected = {
        'a': ['1', 'x'],
        'b': ['2', 'y'],
        'c': ['3', 'z']
    }

    records = textfile.read_separated_lines_with_first_key(file_path, separator='\t')

    assert expected == records


def test_read_separated_keeping_empty_columns():
    file_path = os.path.join(os.path.dirname(__file__), 'empty_column_file.txt')

    expected = [
        ['a', '1', 'x', '', ''],
        ['b', '2', '', 'y'],
        ['c', '', '3', 'z'],
    ]

    records = textfile.read_separated_lines(file_path, separator='\t', keep_empty=True)

    assert expected == records


def test_read_key_value_lines():
    file_path = os.path.join(os.path.dirname(__file__), 'key_value_file.txt')

    expected = {
        'a': '1',
        'b': '2',
        'c': '3'
    }

    records = textfile.read_key_value_lines(file_path, separator=' ')

    assert expected == records


def test_write_separated_lines_sorted():
    data = {
        'hallo-0_103': 'hallo-0_1',
        'hallo-0_122': 'hallo-0',
        'hallo-0_1031': 'hallo-0_1',
        'hallo-0_1322': 'hallo-0',
        'hallo-0_1224': 'hallo-0'
    }

    f, path = tempfile.mkstemp(text=True)
    os.close(f)

    textfile.write_separated_lines(path, data, separator=' ', sort_by_column=1)

    f = open(path, 'r')
    value = f.read()
    f.close()

    lines = value.strip().split('\n')

    assert 5 == len(lines)

    assert lines[0].endswith('hallo-0')
    assert lines[1].endswith('hallo-0')
    assert lines[2].endswith('hallo-0')
    assert lines[3].endswith('hallo-0_1')
    assert lines[4].endswith('hallo-0_1')



import numpy as np

from . import container

SAMPLING_RATE_ATTR = 'sampling-rate'
MAX_INT16_VALUE = np.iinfo(np.int16).max


class AudioContainer(container.Container):
    """
    Container to store raw audio samples.

    Notes:
        The samples are stored  as 16-Bit Integers.
        But all methods expect or return the samples as 32-Bit Floats,
        in the range of -1.0 to 1.0.
    """

    def get(self, key, mem_map=True):
        """
        Return the samples for the given key and the sampling-rate.

        Args:
            key (str): The key to read the data from.
            mem_map (bool): If ``True`` returns the data as
                            memory-mapped array, otherwise a copy is returned.

        Note:
            The container has to be opened in advance.

        Returns:
            tuple: A tuple containing the samples as numpy array
                   with ``np.float32`` [-1.0,1.0] and the sampling-rate.
        """
        self.raise_error_if_not_open()

        if key in self._file:
            data = self._file[key]
            sampling_rate = data.attrs[SAMPLING_RATE_ATTR]

            if not mem_map:
                data = data[()]

            data = np.float32(data) / MAX_INT16_VALUE

            return data, sampling_rate

    def set(self, key, samples, sampling_rate):
        """
        Set the samples and sampling-rate for the given key.
        Existing data will be overwritten.
        The samples have to have ``np.float32`` datatype and values in
        the range of -1.0 and 1.0.

        Args:
            key (str): A key to store the data for.
            samples (numpy.ndarray): 1-D array of audio samples (np.float32).
            sampling_rate (int): The sampling-rate of the audio samples.

        Note:
            The container has to be opened in advance.
        """
        if not np.issubdtype(samples.dtype, np.floating):
            raise ValueError('Samples are required as np.float32!')

        if len(samples.shape) > 1:
            raise ValueError('Only single channel supported!')

        self.raise_error_if_not_open()

        if key in self._file:
            del self._file[key]

        samples = (samples * MAX_INT16_VALUE).astype(np.int16)

        dset = self._file.create_dataset(key, data=samples)
        dset.attrs[SAMPLING_RATE_ATTR] = sampling_rate

    def append(self, key, samples, sampling_rate):
        """
        Append the given samples to the data that already exists
        in the container for the given key.

        Args:
            key (str): A key to store the data for.
            samples (numpy.ndarray): 1-D array of audio samples (int-16).
            sampling_rate (int): The sampling-rate of the audio samples.

        Note:
            The container has to be opened in advance.
            For appending to existing data the HDF5-Dataset has to be chunked,
            so it is not allowed to first add data via ``set``.
        """
        if not np.issubdtype(samples.dtype, np.floating):
            raise ValueError('Samples are required as np.float32!')

        if len(samples.shape) > 1:
            raise ValueError('Only single channel supported!')

        existing = self.get(key, mem_map=True)
        samples = (samples * MAX_INT16_VALUE).astype(np.int16)

        if existing is not None:
            existing_samples, existing_sr = existing

            if existing_sr != sampling_rate:
                raise ValueError('Different sampling-rate than existing data!')

            num_existing = existing_samples.shape[0]
            self._file[key].resize(num_existing + samples.shape[0], 0)
            self._file[key][num_existing:] = samples
        else:
            dset = self._file.create_dataset(key, data=samples,
                                             chunks=True, maxshape=(None,))

            dset.attrs[SAMPLING_RATE_ATTR] = sampling_rate



import os

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus.io import common_voice

from tests import resources
from . import reader_test as rt


class TestCommonVoiceReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('common_voice')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'clips')

    EXPECTED_NUMBER_OF_TRACKS = 9
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('c4b', 'c4b.mp3'),
        rt.ExpFileTrack('8ea', '8ea.mp3'),
        rt.ExpFileTrack('67c', '67c.mp3'),
        rt.ExpFileTrack('f08', 'f08.mp3'),
        rt.ExpFileTrack('b5c', 'b5c.mp3'),
        rt.ExpFileTrack('8f4', '8f4.mp3'),
        rt.ExpFileTrack('7f4', '7f4.mp3'),
        rt.ExpFileTrack('059', '059.mp3'),
        rt.ExpFileTrack('d08', 'd08.mp3'),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 7
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('17e', 1, issuers.Gender.MALE, issuers.AgeGroup.SENIOR, None),
        rt.ExpSpeaker('cb3', 2, issuers.Gender.MALE, issuers.AgeGroup.ADULT, None),
        rt.ExpSpeaker('aa3', 2, issuers.Gender.FEMALE, issuers.AgeGroup.ADULT, None),
        rt.ExpSpeaker('90a', 1, issuers.Gender.MALE, issuers.AgeGroup.ADULT, None),
        rt.ExpSpeaker('b0f', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('5ec', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('72d', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 9
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('c4b', 'c4b', '17e', 0, float('inf')),
        rt.ExpUtterance('8ea', '8ea', 'cb3', 0, float('inf')),
        rt.ExpUtterance('67c', '67c', 'cb3', 0, float('inf')),
        rt.ExpUtterance('f08', 'f08', 'aa3', 0, float('inf')),
        rt.ExpUtterance('b5c', 'b5c', 'aa3', 0, float('inf')),
        rt.ExpUtterance('8f4', '8f4', '90a', 0, float('inf')),
        rt.ExpUtterance('7f4', '7f4', 'b0f', 0, float('inf')),
        rt.ExpUtterance('059', '059', '5ec', 0, float('inf')),
        rt.ExpUtterance('d08', 'd08', '72d', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        'c4b': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        'f08': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        '8f4': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        'd08': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)
        ],
    }

    EXPECTED_LABELS = {
        'c4b': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Man sollte', 0, float('inf')),
        ],
        'f08': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Valentin', 0, float('inf')),
        ],
        '8f4': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Es', 0, float('inf')),
        ],
        '7f4': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Zieht euch', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 3

    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('train', [
            'c4b',
            '8ea',
            '67c',
            'f08',
            'b5c',
        ]),
        rt.ExpSubview('dev', [
            '8f4',
            '7f4',
            '059',
            'd08',
        ]),
        rt.ExpSubview('validated', [
            '8f4',
        ]),
    ]

    def load(self):
        reader = common_voice.CommonVoiceReader()
        return reader.load(self.SAMPLE_PATH)



"""
This module contains functions for working with names.
For example to generate identifiers or find an identifier which not already exists in a given list.
"""

import random
import string


def index_name_if_in_list(name, name_list, suffix='', prefix=''):
    """
    Find a unique name by adding an index to the name so it is unique within the given list.

    Parameters:
        name (str): Name
        name_list (iterable): List of names that the new name must differ from.
        suffix (str): The suffix to append after the index.
        prefix (str): The prefix to append in front of the index.

    Returns:
        str: A unique name within the given list.
    """
    new_name = '{}'.format(name)
    index = 1

    while new_name in name_list:
        new_name = '{}_{}{}{}'.format(name, prefix, index, suffix)
        index += 1

    return new_name


def generate_name(length=15, not_in=None):
    """
    Generates a random string of lowercase letters with the given length.

    Parameters:
        length (int): Length of the string to output.
        not_in (list): Only return a string not in the given iterator.

    Returns:
        str: A new name thats not in the given list.
    """
    value = ''.join(random.choice(string.ascii_lowercase) for i in range(length))

    while (not_in is not None) and (value in not_in):
        value = ''.join(random.choice(string.ascii_lowercase) for i in range(length))

    return value



import numpy as np

from . import base


class Stack(base.Reduction):
    """
    Stack the features from multiple inputs.
    All input matrices have to be of the same length (same number of frames).
    """

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        return np.hstack(chunk.data)



import numpy as np
import librosa

from audiomate import tracks
from audiomate.processing import pipeline

from tests import resources


class TestOnsetStrength:

    def test_compute(self):
        test_file_path = resources.sample_wav_file('wav_1.wav')
        y, sr = librosa.load(test_file_path, sr=None)
        frames = librosa.util.frame(y, frame_length=2048, hop_length=1024).T

        # EXPECTED
        S = np.abs(librosa.stft(y, center=False, n_fft=2048, hop_length=1024)) ** 2
        S = librosa.feature.melspectrogram(S=S, n_mels=128, sr=sr)
        S = librosa.power_to_db(S)
        exp_onsets = librosa.onset.onset_strength(S=S, center=False).T
        exp_onsets = exp_onsets.reshape(exp_onsets.shape[0], 1)

        # ACTUAL
        onset = pipeline.OnsetStrength()
        onsets = onset.process_frames(frames, sr, last=True)

        assert np.allclose(onsets, exp_onsets)

    def test_compute_online(self):
        test_file_path = resources.sample_wav_file('wav_1.wav')
        y, sr = librosa.load(test_file_path, sr=None)

        # EXPECTED
        y_pad = np.pad(y, (0, 1024), mode='constant', constant_values=0)
        S = np.abs(librosa.stft(y_pad, center=False, n_fft=2048, hop_length=1024)) ** 2
        S = librosa.feature.melspectrogram(S=S, n_mels=128, sr=sr)
        S = librosa.power_to_db(S)
        exp_onsets = librosa.onset.onset_strength(S=S, center=False).T
        exp_onsets = exp_onsets.reshape(exp_onsets.shape[0], 1)

        # ACTUAL
        test_file = tracks.FileTrack('idx', test_file_path)
        onset = pipeline.OnsetStrength()
        onset_gen = onset.process_track_online(test_file, 2048, 1024, chunk_size=5)

        chunks = list(onset_gen)
        onsets = np.vstack(chunks)

        print(onsets.shape, exp_onsets.shape)

        assert np.allclose(onsets, exp_onsets)



def length_of_overlap(first_start, first_end, second_start, second_end):
    """
    Find the length of the overlapping part of two segments.

    Args:
        first_start (float): Start of the first segment.
        first_end (float): End of the first segment.
        second_start (float): Start of the second segment.
        second_end (float): End of the second segment.

    Return:
        float: The amount of overlap or 0 if they don't overlap at all.
    """
    if first_end <= second_start or first_start >= second_end:
        return 0.0

    if first_start < second_start:
        if first_end < second_end:
            return abs(first_end - second_start)
        else:
            return abs(second_end - second_start)

    if first_start > second_start:
        if first_end > second_end:
            return abs(second_end - first_start)
        else:
            return abs(first_end - first_start)
